# This file is execfile()d with the current directory set to its containing dir.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import shutil
import sys

# -- Path setup --------------------------------------------------------------

__location__ = os.path.dirname(__file__)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.join(__location__, "../src"))

# -- Run sphinx-apidoc -------------------------------------------------------
# This hack is necessary since RTD does not issue `sphinx-apidoc` before running
# `sphinx-build -b html . _build/html`. See Issue:
# https://github.com/readthedocs/readthedocs.org/issues/1139
# DON'T FORGET: Check the box "Install your project inside a virtualenv using
# setup.py install" in the RTD Advanced Settings.
# Additionally it helps us to avoid running apidoc manually

try:  # for Sphinx >= 1.7
    from sphinx.ext import apidoc
except ImportError:
    from sphinx import apidoc

output_dir = os.path.join(__location__, "api")
module_dir = os.path.join(__location__, "../src/nifigator")
try:
    shutil.rmtree(output_dir)
except FileNotFoundError:
    pass

try:
    import sphinx

    cmd_line = f"sphinx-apidoc --implicit-namespaces -f -o {output_dir} {module_dir} ../src/nifigator/nafdocument.py ../src/nifigator/const.py"

    args = cmd_line.split(" ")
    if tuple(sphinx.__version__.split(".")) >= ("1", "7"):
        # This is a rudimentary parse_version to avoid external dependencies
        args = args[1:]

    apidoc.main(args)
except Exception as e:
    print("Running `sphinx-apidoc` failed!\n{}".format(e))

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "nifigator"
copyright = "2023, Willem Jan Willemse"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# version: The short X.Y version.
# release: The full version, including alpha/beta/rc tags.
# If you don’t need the separation provided between version and release,
# just set them both to the same value.
try:
    from nifigator import __version__ as version
except ImportError:
    version = ""

if not version or version.lower() == "unknown":
    version = os.getenv("READTHEDOCS_VERSION", "unknown")  # automatically set by RTD

release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".venv"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If this is True, todo emits a warning for each TODO entries. The default is False.
todo_emit_warnings = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "pydata_sphinx_theme"

html_context = {"default_mode": "dark"}
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {"sidebar_width": "300px", "page_width": "1200px"}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = ""

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None



# -*- coding: utf-8 -*-

import logging
import uuid
from collections import defaultdict
from typing import Optional, Union
from zipfile import ZipFile

from rdflib import Graph
from rdflib.namespace import DC, DCTERMS, NamespaceManager
from rdflib.store import Store
from rdflib.term import IdentifiedNode, URIRef

from .const import ITSRDF, NIF, OLIA
from .converters import nafConverter
from .nafdocument import NafDocument
from .nifobjects import (
    NifContext,
    NifContextCollection,
)

DEFAULT_URI = "https://mangosaurus.eu/rdf-data/"
DEFAULT_PREFIX = "mangosaurus"


class NifGraph(Graph):

    """
    An NIF Graph

    The constructor accepts the same arguments as a `rdflib.Graph`.

    :param file: name of the file to read

    :param nafdocument: an xml file in NLP Annotation Format

    :param collection: an NifContextCollection

    """

    def __init__(
        self,
        file: str = None,
        nafdocument: NafDocument = None,
        collection: NifContextCollection = None,
        URIScheme: str = None,
        store: Union[Store, str] = "default",
        identifier: Optional[Union[IdentifiedNode, str]] = None,
        namespace_manager: Optional[NamespaceManager] = None,
        base: Optional[str] = None,
        bind_namespaces: str = "core",
    ):
        """
        An NIF Graph

        :param file: name of the file to read

        :param nafdocument: an xml file in NLP Annotation Format

        :param collection: an NifContextCollection

        """

        super(NifGraph, self).__init__(
            store=store,
            identifier=identifier,
            namespace_manager=namespace_manager,
            base=base,
            bind_namespaces=bind_namespaces,
        )

        self.URIScheme = URIScheme

        self.bind("rdf", ITSRDF)
        self.bind("rdfs", ITSRDF)
        self.bind("itsrdf", ITSRDF)
        self.bind("xsd", ITSRDF)
        self.bind("dcterms", DCTERMS)
        self.bind("dc", DC)
        self.bind("nif", NIF)
        self.bind("olia", OLIA)

        self.open(file=file, nafdocument=nafdocument, collection=collection)

    def open(
        self,
        file: str = None,
        nafdocument: NafDocument = None,
        collection: NifContextCollection = None,
    ):
        """
        Read data from multiple sources into current `NifGraph` object.

        :param file: name of the file to read

        :param nafdocument: an xml file in NLP Annotation Format

        :param collection: an NifContextCollection

        :return: None

        """
        if file is not None:
            self.__parse_file(file=file)
        elif nafdocument is not None:
            self.__parse_nafdocument(nafdocument=nafdocument)
        elif collection is not None:
            self.__parse_collection(collection=collection)
        return self

    def __parse_nafdocument(self, nafdocument: NafDocument = None):
        """
        Read data from an xml file in NLP Annotation Format

        :param nafdocument: an xml file in NLP Annotation Format

        :return: None

        """
        logging.info(".. Parsing NafDocument to NifGraph")

        doc_uri = nafdocument.header["public"]["{http://purl.org/dc/elements/1.1/}uri"]
        doc_uuid = "nif-" + str(uuid.uuid3(uuid.NAMESPACE_DNS, doc_uri).hex)

        base_uri = DEFAULT_URI
        base_prefix = DEFAULT_PREFIX

        collection = nafConverter(
            collection_name="collection",
            context_name=doc_uuid,
            nafdocument=nafdocument,
            base_uri=base_uri,
            base_prefix=base_prefix,
            URIScheme=self.URIScheme,
        )

        self.__parse_collection(collection)

    # self.parse_collection(collection)

    def __parse_collection(self, collection: NifContextCollection = None):
        """
        Read data from a NifContextCollection object.

        :param collection: a NifContextCollection

        :return: None

        """
        for r in collection.triples():
            self.add(r)

    def __parse_file(self, file: str = None):
        """
        Read data from a file.

        filename ending with "naf.xml": file is read and parsed as
        an xml file in NLP Annotation Format.
        filename ending with "zip": file is extracted and content
        is parsed.

        :param file: a filename.

        :return: None

        """
        if file is not None:
            if file[-7:].lower() == "naf.xml":
                logging.info(".. Parsing file " + file + "")
                nafdocument = NafDocument().open(file)
                self.__parse_nafdocument(nafdocument=nafdocument)
            else:
                if file[-3:].lower() == "zip":
                    # if zip file then parse all files in zip
                    with ZipFile(file, mode="r") as zipfile:
                        logging.info(".. Reading zip file " + file)
                        for filename in zipfile.namelist():
                            with zipfile.open(filename) as f:
                                logging.info(
                                    ".. Parsing file " + filename + " from zip file"
                                )
                                if filename[-4:].lower() == "hext":
                                    self.parse(data=f.read().decode(), format="hext")
                                elif filename[-3:].lower() == "ttl":
                                    self.parse(data=f.read().decode(), format="turtle")
                                else:
                                    self.parse(data=f.read().decode())
                elif file[-4:].lower() == "hext":
                    # if file ends with .hext then parse as hext file
                    with open(file, encoding="utf-8") as f:
                        logging.info(".. Parsing file " + file + "")
                        self.parse(data=f.read(), format="hext")
                else:
                    # otherwise let rdflib determine format
                    with open(file, encoding="utf-8") as f:
                        logging.info(".. Parsing file " + file + "")
                        self.parse(data=f.read())

    @property
    def collection(self, uri: str = DEFAULT_URI):
        """
        This property constructs and returns a `nif:ContextCollection`
        from the `NifGraph`.
        """

        def query_rdf_type(rdf_type: URIRef = None):

            q = (
                """
            SELECT ?s ?p ?o
            WHERE {
                ?s rdf:type """
                + rdf_type.n3(self.namespace_manager)
                + """ .
                ?s ?p ?o .
            }"""
            )
            results = self.query(q)

            d = defaultdict(dict)
            for result in results:
                idx = result[0]
                col = result[1]
                val = result[2]

                if col == NIF.hasContext:
                    if col in d[idx].keys():
                        d[idx][col].append(val)
                    else:
                        d[idx][col] = [val]
                elif val in OLIA:
                    if col in d[idx].keys():
                        d[idx][col].append(val)
                    else:
                        d[idx][col] = [val]
                else:
                    d[idx][col] = val

            return d

        dict_collections = query_rdf_type(NIF.ContextCollection)
        dict_context = query_rdf_type(NIF.Context)
        logging.info(".. extracting nif statements")
        logging.info(
            ".... found " + str(len(dict_collections.keys())) + " collections."
        )
        logging.info(".... found " + str(len(dict_context.keys())) + " contexts.")

        for collection_uri in dict_collections.keys():
            collection = NifContextCollection(uri=collection_uri)
            for predicate in dict_collections[collection_uri].keys():
                if predicate == NIF.hasContext:
                    for context_uri in dict_collections[collection_uri][predicate]:
                        nif_context = NifContext(URIScheme=self.URIScheme).load(
                            graph=self, uri=context_uri
                        )
                        collection.add_context(context=nif_context)
            return collection
        else:
            collection = NifContextCollection(uri=uri)
            for context_uri in dict_context.keys():
                nif_context = NifContext(URIScheme=self.URIScheme).load(
                    graph=self, uri=context_uri
                )
                collection.add_context(context=nif_context)
            return collection

    # @property
    # def catalog(self):
    #     """
    #     """
    #     # derive the conformsTo from the collection
    #     q = """
    #     SELECT ?s
    #     WHERE {
    #         ?a rdf:type nif:ContextCollection .
    #         ?a dcterms:conformsTo ?s
    #     }"""
    #     qres = self.query(q)
    #     dcterms_conformsTo = [row[0] for row in qres]

    #     # find all context in the graphs with corresponding data
    #     q = """SELECT ?s ?p ?o WHERE { ?s rdf:type nif:Context . ?s ?p ?o . }"""
    #     results = self.query(q)

    #     # construct DataFrame from query results
    #     d = defaultdict(dict)
    #     index = list()
    #     columns = set()
    #     for result in results:
    #         idx = result[0]
    #         col = result[1].n3(self.namespace_manager)
    #         if isinstance(result[2], rdflib.term.Literal):
    #             val = result[2].value
    #         else:
    #             val = result[2]
    #         # if ("dc:" in col or "dcterms:" in col or col == "nif:firstSentence"):
    #         d[idx][col] = val
    #         columns.add(col)
    #         if idx not in index:
    #             index.append(idx)
    #     return d

    # df = pd.DataFrame(
    #     index=index,
    #     columns=list(columns),
    #     data = [[d[idx][col] for col in columns] for idx in index]
    # )
    # df['dcterms:conformsTo'] = [", ".join(dcterms_conformsTo)]*len(df.index)
    # df = df.reindex(sorted(df.columns), axis=1)
    # return df

    # @property
    # def olia_annotations(self):
    #     """
    #     """
    #     df = self.extract(rdf_type="nif:Word",
    #                       predicate="nif:oliaLink").reset_index()
    #     df[1] = True
    #     df = df.pivot_table(index=['index'],
    #                         columns=['nif:oliaLink'], values=True, fill_value=0)
    #     df.index = self.natural_sort(df.index)
    #     return df

    # def extract(self,
    #             rdf_type: str=None,
    #             predicate: str="nif:anchorOf"):
    #     """
    #     """
    #     q = """
    #     SELECT ?a ?s
    #     WHERE {
    #         ?a rdf:type """+rdf_type+""" .
    #         ?a """+predicate+""" ?s .
    #     }"""
    #     qres = self.query(q)
    #     # construct DataFrame from query results
    #     index = [
    #         row[0]
    #         for row in qres
    #     ]
    #     columns = [predicate]
    #     data = [
    #         row[1].value
    #         if isinstance(row[1], rdflib.Literal)
    #         else row[1].n3(self.namespace_manager)
    #         for row in qres
    #     ]
    #     df = pd.DataFrame(
    #         index=index,
    #         columns=columns,
    #         data=data
    #     )
    #     # apply natural sort on indices (because they
    #     # contain offsets without preceding zeros)
    #     df.index = self.natural_sort(df.index)
    #     df.index.name = "index"
    #     return df



# -*- coding: utf-8 -*-

"""
"""

from collections import namedtuple
from rdflib.namespace import Namespace
from rdflib.term import URIRef

OffsetBasedString = URIRef(
    "http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core#OffsetBasedString"
)
RFC5147String = URIRef(
    "http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core#RFC5147String"
)

EntityOccurrence = "EntityOccurrence"
TermOccurrence = "TermOccurrence"

NIF = Namespace("http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core#")
NIF_ONTOLOGY = "http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core/2.1"
OLIA = Namespace("http://purl.org/olia/olia.owl#")
ITSRDF = Namespace("http://www.w3.org/2005/11/its/rdf#")

# Tense
# VerbForm
# Case
# Degree
# NumType
# Voice
# Poss

upos2olia = {
    "SYM": URIRef("http://purl.org/olia/olia.owl#Symbol"),
    "ADJ": URIRef("http://purl.org/olia/olia.owl#Adjective"),
    "X": URIRef("http://purl.org/olia/olia-top.owl#Word"),
    "ADV": URIRef("http://purl.org/olia/olia.owl#Adverb"),
    "PUNCT": URIRef("http://purl.org/olia/olia.owl#Punctuation"),
    "AUX": URIRef("http://purl.org/olia/olia.owl#AuxiliaryVerb"),
    "ADP": URIRef("http://purl.org/olia/olia.owl#Adposition"),
    "NUM": URIRef("http://purl.org/olia/olia.owl#Quantifier"),
    "PROPN": URIRef("http://purl.org/olia/olia.owl#ProperNoun"),
    "INTJ": URIRef("http://purl.org/olia/olia.owl#Interjection"),
    "CONJ": URIRef("http://purl.org/olia/olia.owl#CoordinatingConjunction"),
    # possibly an error in Stanza with CCONJ (does not exist in Olia):
    "CCONJ": URIRef("http://purl.org/olia/olia.owl#CoordinatingConjunction"),
    "DET": URIRef("http://purl.org/olia/olia.owl#Determiner"),
    "PART": URIRef("http://purl.org/olia/olia.owl#Particle"),
    "SCONJ": URIRef("http://purl.org/olia/olia.owl#SubordinatingConjunction"),
    "PRON": URIRef("http://purl.org/olia/olia.owl#Pronoun"),
    "NOUN": URIRef("http://purl.org/olia/olia.owl#CommonNoun"),
    "VERB": URIRef("http://purl.org/olia/olia.owl#Verb"),
}

UD2OLIA_mappings = {
    "Definite": {
        "Com": None,
        "Cons": None,
        "Def": "olia:Definite",
        "Ind": "olia:Indefinite",
        "Spec": None,
    },
    "PronType": {
        "Art": "olia:Article",
        "Dem": "olia:DemonstrativePronoun",
        "Emp": "olia:EmphaticPronoun",
        "Exc": None,
        "Ind": "olia:IndefinitePronoun",
        "Int": "olia:InterrogativePronoun",
        "Neg": None,
        "Prs": "olia:PersonalPronoun",
        "Rcp": "olia:ReciprocalPronoun",
        "Rel": "olia:RelativePronoun",
        "Tot": None,
    },
    "Number": {
        "Coll": None,
        "Count": None,
        "Dual": None,
        "Grpa": None,
        "Grpl": None,
        "Inv": None,
        "Pauc": None,
        "Plur": "olia:Plural",
        "Ptan": None,
        "Sing": "olia:Singular",
        "Tri": None,
    },
    "Person": {
        "0": None,
        "1": "olia:First",
        "2": "olia:Second",
        "3": "olia:Third",
        "4": None,
    },
    "Mood": {
        "Adm": None,
        "Cnd": None,
        "Des": None,
        "Imp": "olia:ImperativeMood",
        "Ind": "olia:IndicativeMood",
        "Irr": None,
        "Jus": None,
        "Nec": None,
        "Opt": "olia:OptativeMood",
        "Pot": None,
        "Prp": None,
        "Qot": None,
        "Sub": "olia:SubjunctiveMood",
    },
    "Tense": {
        "Pres": "olia:Present",
        "Past": "olia:Past",
        "Fut": "olia:Future",
        "Pqp": "olia:PluperfectTense",
    },
    "VerbForm": {
        "Inf": "olia:Infinitive",
        "Fin": "olia:FiniteVerb",
        "Part": "olia:Participle",
        "Past": "olia:Past",
        "Ger": "olia:Gerund",
        "Gdv": "olia:NonFiniteVerb",  # (?)
    },
    "Case": {
        "Nom": "olia:Nominative",
        "Gen": "olia:Genitive",
        "Dat": "olia:DativeCase",
        "Acc": "olia:Accusative",
        "Voc": "olia:VocativeCase",
    },
    "Degree": {
        "Pos": "olia:Positive",
        "Sup": "olia:Superlative",
        "Cmp": "olia:Comparative",
    },
    "NumType": {
        "Card": "olia:CardinalNumber",
        "Ord": "olia:OrdinalNumber",
        "Mult": "olia:MultiplicativeNumeral",
        "Frac": "olia:Fraction",
    },
    "NumForm": {
        "Word": "olia:LetterNumeral",
        "Digit": "olia:DigitNumeral",
        "Roman": "olia:RomanNumeral",
    },
    "Voice": {
        "Pass": "olia:PassiveVoice",
        "Act": "olia:ActiveVoice",
        "Mid": "olia:MiddleVoice",
    },
    "Gender": {
        "Com": "olia:CommonGender",
        "Neut": "olia:Neuter",
        "Com,Neut": "olia:CommonGender",  # correct?
        "Masc,Neut": "olia:CommonGender",  # correct?
        "Fem,Masc": "olia:CommonGender",  # correct?
        "Fem": "olia:Feminine",
        "Masc": "olia:Masculine",
    },
    "pos": {
        "adj": "olia:Adjective",
        "adp": "olia:Adposition",
        "adv": "olia:Adverb",
        "aux": "olia:AuxiliaryVerb",
        "conj": "olia:CoordinatingConjunction",
        "cconj": "olia:CoordinatingConjunction",  # ??
        "det": "olia:Determiner",
        "intj": "olia:Interjection",
        "noun": "olia:CommonNoun",
        "num": "olia:Quantifier",
        "part": "olia:Particle",
        "pron": "olia:Pronoun",
        "propn": "olia:ProperNoun",
        "punct": "olia:Punctuation",
        "sconj": "olia:SubordinatingConjunction",
        "sym": "olia:Symbol",
        "verb": "olia:Verb",
        "x": "olia:X",  # &olia-top;Word"  # not correct
    },
    "Polarity": {
        "Neg": "olia:Negation",
    },
    "ExtPos": {
        "ADP": "olia:Adposition",
        "ADV": "olia:Adverb",
        "CCONJ": "olia:CoordinatingConjunction",  # ??
        "PRON": "olia:Pronoun",
        "SCONJ": "olia:SubordinatingConjunction",
    },
    "Typo": {
        "Yes": "olia:Typo",
    },
    "Style": {
        "Arch": None,
        "Coll": None,
        "Expr": None,
        "Form": "olia:FormalRegister",
        "Rare": None,
        "Slng": "olia:SlangRegister",
        "Vrnc": None,
        "Vulg": "olia:VulgarRegister",
    },
    "Aspect": {"Perf": "olia:PerfectiveAspect", "Imp": "olia:ImperfectiveAspect"},
    "Reflex": {"Yes": "olia:ReflexivePronoun"},
    "Abbr": {"Yes": "olia:Abbreviation"},
    "Poss": {"Yes": "olia:PossessivePronoun"},
    "Foreign": {"Yes": "olia:Foreign"},
}


def mapobject(p: str = "", o: str = ""):
    if p not in UD2OLIA_mappings.keys():
        print("UD Not found: " + p)
    else:
        if o not in UD2OLIA_mappings[p].keys():
            print("UD Not found: " + p + " , " + o)
    return (
        UD2OLIA_mappings.get(p, {})
        .get(o, "olia:" + o)
        .replace("olia:", "http://purl.org/olia/olia.owl#")
    )


# nafdocument definition

ProcessorElement = namedtuple(
    "lp", "name version model timestamp beginTimestamp endTimestamp hostname"
)

WordformElement = namedtuple("WfElement", "id sent para page offset length xpath text")

TermElement = namedtuple(
    "TermElement",
    "id type lemma pos morphofeat netype case head component_of \
    compound_type span ext_refs comment",
)

Entity = namedtuple("Entity", "start end type")

EntityElement = namedtuple(
    "EntityElement", "id type status source span ext_refs comment"
)

DependencyRelation = namedtuple(
    "DependencyRelation", "from_term to_term rfunc case comment"
)

ChunkElement = namedtuple("ChunkElement", "id head phrase case span comment")

RawElement = namedtuple("RawElement", "text")

MultiwordElement = namedtuple(
    "MultiwordElement", "id lemma pos morphofeat case status type components"
)

ComponentElement = namedtuple(
    "ComponentElement", "id type lemma pos morphofeat netype case head span"
)



# -*- coding: utf-8 -*-

import logging
from collections import OrderedDict, deque
from typing import Union

import iribaker
from rdflib import Graph
from rdflib.namespace import DC, DCTERMS, RDF, XSD
from rdflib.term import Literal, URIRef

from .const import (
    ITSRDF,
    NIF,
    NIF_ONTOLOGY,
    OLIA,
    EntityOccurrence,
    OffsetBasedString,
    RFC5147String,
    TermOccurrence,
    mapobject,
    upos2olia,
)
from .utils import delete_accents, delete_diacritics, natural_sort


class NifContext:
    pass


class NifSentence:
    pass


class NifPhrase:
    pass


class NifWord:
    pass


class NifBase(object):
    """
    A NIF Base

    :param uri: the uri of the object

    """

    def __init__(self, uri: Union[URIRef, str] = None):
        self.set_uri(uri)

    def __eq__(self, other):
        return self._uri == other._uri

    @property
    def uri(self):
        """
        Returns the uri of the object
        """
        if self._uri is not None:
            return self._uri
        else:
            return None

    def set_uri(self, uri: Union[URIRef, str] = None):
        """
        Sets the uri of the object.
        If the uri is a string then it is converted to an iri.
        """
        if isinstance(uri, str):
            self._uri = URIRef(iribaker.to_iri(uri))
        else:
            self._uri = uri


class NifString(NifBase):
    """
    A NIF String

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
    ):
        self.set_URIScheme(URIScheme)
        self.set_beginIndex(beginIndex)
        self.set_endIndex(endIndex)
        self.set_referenceContext(referenceContext)
        self.set_uri(uri)

    def __eq__(self, other):
        return (
            (self._URIScheme == other._URIScheme)
            & (self._beginIndex == other._beginIndex)
            & (self._endIndex == other._endIndex)
            & (self._referenceContext.uri == other._referenceContext.uri)
            & super(NifBase, self).__eq__(other)
        )

    @property
    def beginIndex(self):
        """
        Returns the start index of the context string as an `int`.
        """
        if self._beginIndex is not None:
            return int(self._beginIndex)
        else:
            return None

    @property
    def endIndex(self):
        """
        Returns the end index of the context string as an `int`.
        """
        if self._endIndex is not None:
            return int(self._endIndex)
        else:
            return None

    @property
    def referenceContext(self):
        """
        Returns the context of the current object
        """
        if self._referenceContext is not None:
            return self._referenceContext
        else:
            return None

    @property
    def anchorOf(self):
        """
        Returns the string of the object as a `str`.
        """
        if self._anchorOf is not None:
            return self._anchorOf.value
        else:
            return None

    @property
    def anchorOf_no_accents(self):
        """
        Returns the string without accents of the object as a `str`.
        """
        if self._anchorOf is not None:
            if (
                self._referenceContext is not None
                and self._referenceContext._metadata is not None
            ):
                lang = self._referenceContext._metadata.get(
                    DC.language, Literal("en", datatype=XSD.string)
                )
            else:
                lang = Literal("en", datatype=XSD.string)
            return Literal(
                delete_accents(s=self._anchorOf.value, lang=lang), datatype=XSD.string
            )
        else:
            return None

    @property
    def anchorOf_no_diacritics(self):
        """
        Returns the string without diacritics of the object as a `str`.
        """
        if self._anchorOf is not None:
            if (
                self._referenceContext is not None
                and self._referenceContext._metadata is not None
            ):
                lang = self._referenceContext._metadata.get(
                    DC.language, Literal("en", datatype=XSD.string)
                )
            else:
                lang = Literal("en", datatype=XSD.string)
            return Literal(
                delete_diacritics(s=self._anchorOf.value, lang=lang),
                datatype=XSD.string,
            )
        else:
            return None

    @property
    def URIScheme(self):
        """
        Returns the URIScheme
        """
        return self._URIScheme

    def set_uri(self, uri: Union[URIRef, str] = None):
        """
        Sets the uri of the object
        """
        if uri is not None:
            if not isinstance(uri, URIRef):
                if not isinstance(self, NifContext):
                    uri_format = OffsetBasedString
                    if uri_format == OffsetBasedString:
                        uri = (
                            uri
                            + "#offset_"
                            + str(self.beginIndex)
                            + "_"
                            + str(self.endIndex)
                        )
                    elif uri_format == RFC5147String:
                        uri = (
                            uri
                            + "#char="
                            + str(self.beginIndex)
                            + ","
                            + str(self.endIndex)
                        )
            super().set_uri(uri=uri)

    def set_beginIndex(self, beginIndex: Union[Literal, int] = None):
        """
        Sets the start of the index of the string.
        The type of beginIndex can be a `Literal` or an `int`.
        If the type is an `int` then it is converted to a Literal.
        """
        if isinstance(beginIndex, int):
            self._beginIndex = Literal(beginIndex, datatype=XSD.nonNegativeInteger)
        else:
            self._beginIndex = beginIndex

    def set_endIndex(self, endIndex: Union[Literal, int] = None):
        """
        Sets the end of the index of the string.
        The type of endIndex can be a `Literal` or an `int`.
        If the type is an `int` then it is converted to a Literal.
        """
        if isinstance(endIndex, int):
            self._endIndex = Literal(endIndex, datatype=XSD.nonNegativeInteger)
        else:
            self._endIndex = endIndex

    def set_referenceContext(self, referenceContext: NifContext = None):
        """
        Sets the referenceContext of the object.
        The anchorOf is set to the referenceContext string
        with the beginIndex and endIndex.
        """
        if referenceContext is not None:
            if referenceContext.isString is not None:
                anchorOf = referenceContext.isString[self.beginIndex : self.endIndex]
                self.set_anchorOf(anchorOf)
            self._referenceContext = referenceContext

    def set_anchorOf(self, anchorOf: str = None):
        """
        Sets the anchorOf of the object.
        The anchorOf should be consistent with the string in
        the referenceContext.
        """
        if isinstance(anchorOf, str):
            self._anchorOf = Literal(anchorOf, datatype=XSD.string)
        elif isinstance(anchorOf, Literal):
            self._anchorOf = anchorOf

    def set_URIScheme(self, URIScheme: str = None):
        """
        Sets the URIScheme of the object
        """
        self._URIScheme = URIScheme

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            if self._URIScheme == OffsetBasedString:
                yield (self.uri, RDF.type, NIF.OffsetBasedString)
            elif self._URIScheme == RFC5147String:
                yield (self.uri, RDF.type, NIF.RFC5147String)
            yield (self.uri, RDF.type, NIF.String)
            if self._beginIndex is not None:
                yield (self.uri, NIF.beginIndex, self._beginIndex)
            if self._endIndex is not None:
                yield (self.uri, NIF.endIndex, self._endIndex)
            if self._referenceContext is not None:
                yield (self.uri, NIF.referenceContext, self._referenceContext.uri)


class NifContext(NifString):
    """
    A NIF Context

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param sourceUrl: the source url of the context

    :param predLang: the predominant language of the context

    :param isString: the string of the context

    :param metadata: a list of URIRefs with metadata

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        sourceUrl: URIRef = None,
        predLang: URIRef = None,
        isString: Union[Literal, str] = None,
        metadata: dict = None,
    ):
        self.set_sourceUrl(sourceUrl)
        self.set_predLang(predLang)
        self.set_isString(isString)
        self.set_metadata(metadata)
        self.set_Sentences(None)
        self.set_Paragraphs(None)
        self.set_Pages(None)
        self.set_Phrases(None)
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=0 if isString is not None else None,
            endIndex=len(isString) if isString is not None else None,
            referenceContext=self,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:Context) uri = {self.uri.n3()}\n"
        if self.sourceUrl is not None:
            s += f"  sourceUrl : {self.sourceUrl.n3()}\n"
        if self.predLang is not None:
            s += f"  predLang : {self.predLang.n3()}\n"
        if self.isString is not None:
            if len(self.isString) > 1000:
                s += f'  isString : {repr(self.isString[0:1000]+"... ")}\n'
            else:
                s += f"  isString : {repr(self.isString)}\n"
        if self.firstSentence is not None:
            s += f"  firstSentence : {repr(self.firstSentence.anchorOf)}\n"
        if self.lastSentence is not None:
            s += f"  lastSentence : {repr(self.lastSentence.anchorOf)}\n"
        if self._metadata is not None and self._metadata != {}:
            for d in self._metadata.keys():
                s += f"  {d} : {self._metadata[d]}\n"
        return s

    def __eq__(self, other):
        return (
            (self._URIScheme == other._URIScheme)
            & (self._uri == other._uri)
            & (self.sourceUrl == other.sourceUrl)
            & (self.predLang == other.predLang)
            & (self.isString == other.isString)
            & (self.metadata == other.metadata)
            & super(NifBase, self).__eq__(other)
        )

    @property
    def sourceUrl(self):
        """
        Returns the sourceUrl of the context
        """
        if self._sourceUrl is not None:
            return self._sourceUrl
        else:
            return None

    @property
    def predLang(self):
        """
        Returns the predLang of the context
        """
        if self._predLang is not None:
            return self._predLang
        else:
            return None

    @property
    def isString(self):
        """
        Returns the isString of the context
        """
        if self._isString is not None:
            return self._isString.value
        else:
            return None

    @property
    def firstSentence(self):
        """
        Returns the first sentence of the context.
        """
        if self._sentences is not None:
            return self._sentences[0]
        else:
            return None

    @property
    def lastSentence(self):
        """
        Returns the last sentence of the context.
        """
        if self._sentences is not None:
            return self._sentences[-1]
        else:
            return None

    @property
    def sentences(self):
        """
        Returns all sentences in the context as a list.
        """
        if self._sentences is not None:
            return list(self._sentences)
        else:
            return None

    @property
    def firstParagraph(self):
        """
        Returns the first paragraph of the context.
        """
        if self._paragraphs is not None:
            return self._paragraphs[0]
        else:
            return None

    @property
    def lastParagraph(self):
        """
        Returns the last paragraph of the context.
        """
        if self._paragraphs is not None:
            return self._paragraphs[-1]
        else:
            return None

    @property
    def paragraphs(self):
        """
        Returns all paragraphs in the context as a list.
        """
        if self._paragraphs is not None:
            return list(self._paragraphs)
        else:
            return None

    @property
    def firstPage(self):
        """
        Returns the first page of the context.
        """
        if self._pages is not None:
            return self._pages[0]
        else:
            return None

    @property
    def lastPage(self):
        """
        Returns the last page of the context.
        """
        if self._pages is not None:
            return self._pages[-1]
        else:
            return None

    @property
    def pages(self):
        """
        Returns all pages in the context as a list.
        """
        if self._pages is not None:
            return list(self._pages)
        else:
            return None

    @property
    def firstPhrase(self):
        """
        Returns the first phrase of the context.
        """
        if self._phrases is not None:
            return self._phrases[0]
        else:
            return None

    @property
    def lastPhrase(self):
        """
        Returns the last phrase of the context.
        """
        if self._phrases is not None:
            return self._phrases[-1]
        else:
            return None

    @property
    def phrases(self):
        """
        Returns all phrases in the context as a list.
        """
        if self._phrases is not None:
            return list(self._phrases)
        else:
            return None

    def set_metadata(self, metadata: dict = None):
        """
        Sets the metadata of the context (a dict of predicates and objects)
        """
        if metadata is not None:
            self._metadata = metadata
        else:
            self._metadata = {}

    def set_sourceUrl(self, sourceUrl: URIRef = None):
        """
        Sets the sourceUrl of the context
        """
        self._sourceUrl = sourceUrl

    def set_predLang(self, predLang: Union[URIRef, str] = None):
        """
        Sets the predominant language of the context
        """
        if predLang is not None:
            if not isinstance(predLang, URIRef):
                self._predLang = Literal(predLang)
            else:
                self._predLang = predLang
        else:
            self._predLang = None

    def set_isString(self, isString: Union[Literal, str] = None):
        """
        Sets the string of the context (rdflib.Literal or string)
        """
        if isinstance(isString, str):
            self._isString = Literal(isString, datatype=XSD.string)
        else:
            self._isString = isString

    def set_Pages(self, pages: list = None):
        """
        Sets the pages of the context (a list of NifPage)
        """
        if pages is not None and pages != []:
            self._pages = deque(pages)
        else:
            self._pages = None

    def set_Paragraphs(self, paragraphs: list = None):
        """
        Sets the paragraphs of the context (a list of NifParagraph)
        """
        if paragraphs is not None and paragraphs != []:
            self._paragraphs = deque(paragraphs)
        else:
            self._paragraphs = None

    def set_Phrases(self, phrases: list = None):
        """
        Sets the phrases of the context (a list of NifPhrases)
        """
        if phrases is not None and phrases != []:
            self._phrases = deque(phrases)
        else:
            self._phrases = None

    def set_Sentences(self, sentences: list = None):
        """
        Sets the sentences of the context (a list of NifSentence)
        """
        if sentences is not None and sentences != []:
            self._sentences = deque(sentences)
        else:
            self._sentences = None

    def add_sentence(self, sentence: NifSentence = None):
        """
        Adds a sentences to the context (a NifSentence)
        """
        if sentence is not None:
            if self._sentences is None:
                self._sentences = deque([sentence])
            else:
                self._sentences.append(sentence)

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:

            yield (self.uri, RDF.type, NIF.Context)
            for key in self._metadata.keys():
                yield (self.uri, key, self._metadata[key])
            if self._isString is not None:
                yield (self.uri, NIF.isString, self._isString)
            if self._sourceUrl is not None:
                yield (self.uri, NIF.sourceUrl, self._sourceUrl)
            if self._predLang is not None:
                yield (self.uri, NIF.predLang, self._predLang)
            if self.firstSentence is not None:
                yield (self.uri, NIF.firstSentence, self.firstSentence.uri)
            if self.lastSentence is not None:
                yield (self.uri, NIF.lastSentence, self.lastSentence.uri)

            if self._sentences is not None:
                for sentence in self._sentences:
                    sentence.triples()

            if self._paragraphs is not None:
                for paragraph in self._paragraphs:
                    paragraph.triples()

            if self._pages is not None:
                for page in self._pages:
                    page.triples()

            if self._phrases is not None:
                for phrase in self._phrases:
                    phrase.triples()

            super().triples()

    def load(self, graph: Graph = None, uri: URIRef = None):
        """
        Load a context from a graph (based on uri)
        """
        metadata = {}
        for s, p, o in graph.triples((uri, RDF.type, None)):
            if o == NIF.OffsetBasedString:
                self.set_URIScheme(o)
        for s, p, o in graph.triples((uri, NIF.isString, None)):
            self.set_isString(o)
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.sourceUrl, None)):
            self.set_sourceUrl(o)
        for s, p, o in graph.triples((uri, NIF.predLang, None)):
            self.set_predLang(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            self.set_anchorOf(o)
        for s, p, o in graph.triples((uri, NIF.firstSentence, None)):
            first_sentence = o
        for s, p, o in graph.triples((uri, NIF.lastSentence, None)):
            last_sentence = o
        for s, p, o in graph.triples((uri, NIF.nextSentence, None)):
            self.set_nextSentence(o)
        for s, p, o in graph.triples((uri, NIF.previousSentence, None)):
            self.set_previousSentence(o)
        for s, p, o in graph.triples((uri, None, None)):
            if p in DC or p in DCTERMS:
                metadata[p] = o
        self.set_uri(uri)
        self.set_referenceContext(self)
        self.set_metadata(metadata)

        page_uris = []
        para_uris = []
        sent_uris = []
        phrase_uris = []
        for ref_uri, _, _ in graph.triples((None, NIF.referenceContext, uri)):
            if (ref_uri, RDF.type, NIF.Page) in graph:
                page_uris.append(ref_uri)
            if (ref_uri, RDF.type, NIF.Paragraph) in graph:
                para_uris.append(ref_uri)
            if (ref_uri, RDF.type, NIF.Sentence) in graph:
                sent_uris.append(ref_uri)
            if (ref_uri, RDF.type, NIF.Phrase) in graph:
                phrase_uris.append(ref_uri)
        page_uris = natural_sort(page_uris)
        para_uris = natural_sort(para_uris)
        sent_uris = natural_sort(sent_uris)
        phrase_uris = natural_sort(phrase_uris)

        # extract pages from graph
        self.set_Pages(
            list(
                NifPage(URIScheme=self.URIScheme).load(
                    graph=graph, referenceContext=self, uri=page_uri
                )
                for page_uri in page_uris
            )
        )
        # extract paragraphs from graph
        self.set_Paragraphs(
            list(
                NifParagraph(URIScheme=self.URIScheme).load(
                    graph=graph, referenceContext=self, uri=para_uri
                )
                for para_uri in para_uris
            )
        )
        # extract phrases from graph
        self.set_Phrases(
            list(
                NifPhrase(URIScheme=self.URIScheme).load(
                    graph=graph, referenceContext=self, uri=phrase_uri
                )
                for phrase_uri in phrase_uris
            )
        )
        # extract sentences from graph
        self.set_Sentences(
            list(
                NifSentence(URIScheme=self.URIScheme).load(
                    graph=graph, referenceContext=self, uri=sent_uri
                )
                for sent_uri in sent_uris
            )
        )
        sentences = self.sentences
        if sentences is not None:

            for sent_idx, sentence in enumerate(sentences):
                if sent_idx > 0:
                    sentence.set_previousSentence(sentences[sent_idx - 1])
                if sent_idx < len(sentences) - 1:
                    sentence.set_nextSentence(sentences[sent_idx + 1])

            if self.firstSentence.uri != first_sentence:
                logging.error(
                    "Error: firstSentence in rdf data not first sentence in context."
                )
            if self.lastSentence.uri != last_sentence:
                logging.error(
                    "Error: lastSentence in rdf data not last sentence in context."
                )

        return self

    def load_from_dict(self, stanza_dict: list = None):
        """
        Load a context from stanza dictionary
        """
        if stanza_dict is not None:

            for sent_idx, sent in enumerate(stanza_dict):

                nif_sent = NifSentence(
                    uri=str(self.uri),
                    beginIndex=sent[0]["start_char"],
                    endIndex=sent[-1]["end_char"],
                    referenceContext=self,
                    URIScheme=self.URIScheme,
                )
                self.add_sentence(nif_sent)

                for word_idx, word in enumerate(sent):
                    nif_word = NifWord(
                        uri=str(self.uri),
                        beginIndex=word["start_char"],
                        endIndex=word["end_char"],
                        referenceContext=self,
                        nifsentence=nif_sent,
                        URIScheme=self.URIScheme,
                    )
                    nif_sent.add_word(nif_word)

                    nif_word.set_lemma(word.get("lemma", None))

                    upos = word.get("upos", None)
                    if upos is not None:
                        if upos in upos2olia.keys():
                            nif_word.set_pos(upos2olia.get(word["upos"]))
                        else:
                            logging.error(
                                ".. part-of-speech tag not found: " + word["upos"]
                            )
                    feats = word.get("feats", None)
                    if feats is not None:
                        for i in feats.split("|"):
                            p = i.split("=")[0]
                            o = i.split("=")[1]
                            olia = mapobject(p, o)
                            if olia is not None:
                                nif_word.set_morphofeats(URIRef(olia))

                for word_idx, word in enumerate(sent):

                    nif_sent._words[word_idx].set_dependencyRelationType(
                        word.get("deprel", None)
                    )

                    dep = word.get("head", None)
                    if dep is not None:
                        if dep != 0:  # if dep is 0 then it is the root
                            nif_sent._words[word_idx].set_dependency(
                                [nif_sent._words[dep - 1]]
                            )

                words = nif_sent._words
                if words is not None:
                    for word_idx, word in enumerate(words):
                        if word_idx < len(words) - 1:
                            word.set_nextWord(words[word_idx + 1])
                        if word_idx > 0:
                            word.set_previousWord(words[word_idx - 1])

            sentences = self.sentences
            if sentences is not None:
                for sent_idx, sentence in enumerate(sentences):
                    if sent_idx < len(sentences) - 1:
                        sentence.set_nextSentence(sentences[sent_idx + 1])
                    if sent_idx > 0:
                        sentence.set_previousSentence(sentences[sent_idx - 1])


class NifStructure(NifString):
    """
    A NIF Structure

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    """

    def __init__(
        self,
        uri: Union[URIRef, str] = None,
        URIScheme: str = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
    ):
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            super().triples()


class NifPhrase(NifStructure):
    """
    A NIF Phrase

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    :param taIdentRef: text analysis identifier reference

    :param taClassRef: text analysis class reference

    :param taConfidence: confidence of the annotation

    :param PhraseType: type of phrase (EntityOccurrence, TermOccurrence)
    """

    def __init__(
        self,
        uri: Union[URIRef, str] = None,
        URIScheme: str = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
        taIdentRef: URIRef = None,
        taClassRef: URIRef = None,
        taConfidence: Union[Literal, float] = None,
        PhraseType: str = None,
        nextPhrase: NifPhrase = None,
        previousPhrase: NifPhrase = None,
    ):
        self.set_taIdentRef(taIdentRef)
        self.set_taClassRef(taClassRef)
        self.set_taConfidence(taConfidence)
        self.set_PhraseType(PhraseType)
        self._set_nextPhrase(nextPhrase)
        self._set_previousPhrase(previousPhrase)
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self._PhraseType == EntityOccurrence:
            s = f"(nif:EntityOccurrence) uri = {self.uri.n3()}\n"
        else:
            s = f"(nif:TermOccurrence) uri = {self.uri.n3()}\n"
        if self.referenceContext is not None:
            s += f"  referenceContext : {self.referenceContext.uri}\n"
        if self.beginIndex is not None:
            s += f"  beginIndex : {self.beginIndex}\n"
        if self.endIndex is not None:
            s += f"  endIndex : {self.endIndex}\n"
        if self.anchorOf is not None:
            s += f'  anchorOf : "{self.anchorOf}"\n'
        if self.taIdentRef is not None:
            s += f"  taIdentRef : {self.taIdentRef}\n"
        if self.taClassRef is not None:
            s += f"  taClassRef : {self.taClassRef}\n"
        if self.taConfidence is not None:
            s += f"  taConfidence : {self.taConfidence}\n"
        return s

    @property
    def PhraseType(self):
        """
        Returns the phrasetype (entity or term occurrence)
        """
        if self._PhraseType is not None:
            return self._PhraseType
        else:
            return None

    @property
    def taIdentRef(self):
        """
        Returns text analysis identifier reference
        """
        if self._taIdentRef is not None:
            return self._taIdentRef
        else:
            return None

    @property
    def taClassRef(self):
        """
        Returns text analysis class reference
        """
        if self._taClassRef is not None:
            return self._taClassRef
        else:
            return None

    @property
    def taConfidence(self):
        """
        Returns text analysis confidence
        """
        if self._taConfidence is not None:
            return float(self._taConfidence.value)
        else:
            return None

    @property
    def nextPhrase(self):
        """
        Returns the next phrase
        """
        if self._nextPhrase is not None:
            return self._nextPhrase
        else:
            return None

    @property
    def previousPhrase(self):
        """
        Returns the previous phrase
        """
        if self._previousPhrase is not None:
            return self._previousPhrase
        else:
            return None

    def set_PhraseType(self, PhraseType: str = None):
        """
        Sets the phrase type (EntityOccurrence or TermOccurrence)
        """
        self._PhraseType = PhraseType

    def set_taIdentRef(self, taIdentRef: Union[URIRef, str] = None):
        """
        Sets the text analysis identifier reference (as a rdflib.URIRef)
        """
        if isinstance(taIdentRef, str):
            self._taIdentRef = URIRef(taIdentRef)
        else:
            self._taIdentRef = taIdentRef

    def set_taClassRef(self, taClassRef: Union[URIRef, str] = None):
        """
        Sets the text analysis class reference (as a rdflib.URIRef)
        """
        if isinstance(taClassRef, str):
            self._taClassRef = URIRef(taClassRef)
        else:
            self._taClassRef = taClassRef

    def set_taConfidence(self, taConfidence: Union[Literal, float] = None):
        """
        Sets the text analysis confidence (float)
        """
        if isinstance(taConfidence, float):
            self._taConfidence = Literal(taConfidence, datatype=XSD.float)
        else:
            self._taConfidence = taConfidence

    def _set_nextPhrase(self, nextPhrase: NifPhrase = None):
        """
        Sets the next phrase
        """
        self._nextPhrase = nextPhrase

    def _set_previousPhrase(self, previousPhrase: NifPhrase = None):
        """
        Sets the previous phrase
        """
        self._previousPhrase = previousPhrase

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.Phrase)
            if self._PhraseType == EntityOccurrence:
                yield (self.uri, RDF.type, NIF.EntityOccurrence)
            elif self._PhraseType == TermOccurrence:
                yield (self.uri, RDF.type, NIF.TermOccurrence)
            if self.taClassRef is not None:
                yield (self.uri, ITSRDF.taClassRef, self._taClassRef)
            if self.taIdentRef is not None:
                yield (self.uri, ITSRDF.taIdentRef, self._taIdentRef)
            if self.taConfidence is not None:
                yield (self.uri, ITSRDF.taConfidence, self._taConfidence)
            super().triples()

    def load(
        self,
        graph: Graph = None,
        referenceContext: NifContext = None,
        uri: URIRef = None,
    ):
        """
        Load a phrase from a graph (based on uri)
        """
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            self.set_anchorOf(o)
        for s, p, o in graph.triples((uri, ITSRDF.taIdentRef, None)):
            self.set_taIdentRef(o)
        for s, p, o in graph.triples((uri, ITSRDF.taClassRef, None)):
            self.set_taClassRef(o)
        for s, p, o in graph.triples((uri, ITSRDF.taConfidence, None)):
            self.set_taConfidence(o)
        for s, p, o in graph.triples((uri, RDF.type, None)):
            if o == EntityOccurrence or o == TermOccurrence:
                self.set_PhraseType(o)
        self.set_uri(uri)
        self.set_referenceContext(referenceContext)

        return self


class NifSentence(NifStructure):
    """
    A NIF Sentence

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    :param nextSentence: the next sentence in the context

    :param previousSentence: the previous sentence in the context

    """

    def __init__(
        self,
        uri: Union[URIRef, str] = None,
        URIScheme: str = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
        nextSentence: Union[URIRef, str] = None,
        previousSentence: Union[URIRef, str] = None,
    ):
        self.set_nextSentence(nextSentence)
        self.set_previousSentence(previousSentence)
        self.set_Words(None)
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:Sentence) uri = {self.uri}\n"
        if self.referenceContext is not None:
            s += f"  referenceContext : {self.referenceContext.uri}\n"
        if self.beginIndex is not None:
            s += f"  beginIndex : {self.beginIndex}\n"
        if self.endIndex is not None:
            s += f"  endIndex : {self.endIndex}\n"
        if self.anchorOf is not None:
            s += f"  anchorOf : {repr(self.anchorOf)}\n"
        if self.nextSentence is not None:
            if len(self.nextSentence.anchorOf) > 100:
                s += f"  nextSentence : {repr(self.nextSentence.anchorOf[0:100])}...\n"
            else:
                s += f"  nextSentence : {repr(self.nextSentence.anchorOf)}\n"
        if self.previousSentence is not None:
            if len(self.previousSentence.anchorOf) > 100:
                ps = repr(self.previousSentence.anchorOf[0:100])
                s += f"  previousSentence : {ps}... \n"
            else:
                ps = repr(self.previousSentence.anchorOf)
                s += f"  previousSentence : {ps}\n"
        if self.firstWord is not None:
            s += f'  firstWord : "{self.firstWord.anchorOf}"\n'
        if self.lastWord is not None:
            s += f'  lastWord : "{self.lastWord.anchorOf}"\n'
        return s

    @property
    def nextSentence(self):
        if self._nextSentence is not None:
            return self._nextSentence
        else:
            return None

    @property
    def previousSentence(self):
        if self._previousSentence is not None:
            return self._previousSentence
        else:
            return None

    @property
    def firstWord(self):
        if self._words is not None:
            return self._words[0]
        else:
            return None

    @property
    def lastWord(self):
        if self._words is not None:
            return self._words[-1]
        else:
            return None

    @property
    def words(self):
        if self._words is not None:
            return list(self._words)
        else:
            return None

    @property
    def lemmas(self):
        return " ".join([w.lemma for w in self.words])

    def set_nextSentence(self, nextSentence: NifSentence = None):
        self._nextSentence = nextSentence

    def set_previousSentence(self, previousSentence: NifSentence = None):
        self._previousSentence = previousSentence

    def set_Words(self, words: list = None):
        if words is not None and words != []:
            self._words = deque(words)
        else:
            self._words = None

    def add_word(self, word: NifWord = None):
        if word is not None:
            if self._words is None:
                self._words = deque([word])
            else:
                self._words.append(word)

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.Sentence)
            super().triples()
            if self.nextSentence is not None:
                yield (self.uri, NIF.nextSentence, self.nextSentence.uri)
            if self.previousSentence is not None:
                yield (self.uri, NIF.previousSentence, self.previousSentence.uri)
            if self.firstWord is not None:
                yield (self.uri, NIF.firstWord, self.firstWord.uri)
            if self.lastWord is not None:
                yield (self.uri, NIF.lastWord, self.lastWord.uri)

            for word in self._words:
                word.triples()

    def load(
        self,
        graph: Graph = None,
        referenceContext: NifContext = None,
        uri: URIRef = None,
    ):
        """
        Load a sentence from a graph (based on uri)
        """
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            self.set_anchorOf(o)
        for s, p, o in graph.triples((uri, NIF.nextSentence, None)):
            self.set_nextSentence(o)
        for s, p, o in graph.triples((uri, NIF.previousSentence, None)):
            self.set_previousSentence(o)
        for s, p, o in graph.triples((uri, NIF.firstWord, None)):
            first_word = o
        for s, p, o in graph.triples((uri, NIF.lastWord, None)):
            last_word = o
        self.set_uri(uri)
        self.set_referenceContext(referenceContext)

        word_uris = []
        for ref_uri, _, _ in graph.triples((None, NIF.sentence, self.uri)):
            if (ref_uri, RDF.type, NIF.Word) in graph:
                word_uris.append(ref_uri)
        word_uris = natural_sort(word_uris)

        # extract words from graph
        words = OrderedDict()
        for word_uri in word_uris:
            words[word_uri] = NifWord(URIScheme=self.URIScheme).load(
                graph=graph,
                referenceContext=self.referenceContext,
                nifsentence=self,
                uri=word_uri,
            )
        self.set_Words(words.values())

        # replace dependency uris by word objects
        for word_idx, word in enumerate(words.values()):
            word.set_dependency([words[dep] for dep in word.dependency])

        words = self.words
        if words is not None:

            # replace nextWord and previousWord uris by word objects
            for word_idx, word in enumerate(words):
                if word_idx > 0:
                    word.set_previousWord(words[word_idx - 1])
                if word_idx < len(words) - 1:
                    word.set_nextWord(words[word_idx + 1])

            # check is firstWord in rdf data is first word in sentence object
            if self.firstWord.uri != first_word:
                logging.error(
                    "Error: firstWord in rdf data not first word in sentence."
                )

            # check is lastWord in rdf data is last word in sentence object
            if self.lastWord.uri != last_word:
                logging.error("Error: lastWord in rdf data not last word in context.")

        return self


class NifParagraph(NifStructure):
    """
    A NIF Paragraph

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
    ):
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:Paragraph) uri = {self.uri}\n"
        if self.beginIndex is not None:
            s += f"  beginIndex : {self.beginIndex}\n"
        if self.endIndex is not None:
            s += f"  endIndex : {self.endIndex}\n"
        if self.anchorOf is not None:
            if len(self.anchorOf) > 1000:
                s += f'  anchorOf : {repr(self.anchorOf[0:1000]+"... ")}\n'
            else:
                s += f"  anchorOf : {repr(self.anchorOf)}\n"
        return s

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.Paragraph)
            super().triples()

    def load(
        self,
        graph: Graph = None,
        referenceContext: NifContext = None,
        uri: URIRef = None,
    ):
        """
        Load a paragraph from a graph (based on uri)
        """
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            self.set_anchorOf(o)
        self.set_uri(uri)
        self.set_referenceContext(referenceContext)
        return self


class NifPage(NifStructure):
    """
    A NIF Page

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
    ):
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:Page) uri = {self.uri}\n"
        if self.beginIndex is not None:
            s += f"  beginIndex : {self.beginIndex}\n"
        if self.endIndex is not None:
            s += f"  endIndex : {self.endIndex}\n"
        if self.anchorOf is not None:
            if len(self.anchorOf) > 1000:
                s += f'  anchorOf : {repr(self.anchorOf[0:1000]+"... ")}\n'
            else:
                s += f"  anchorOf : {repr(self.anchorOf)}\n"
        return s

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.Page)
            super().triples()

    def load(
        self,
        graph: Graph = None,
        referenceContext: NifContext = None,
        uri: URIRef = None,
    ):
        """
        Load a page from a graph (based on uri)
        """
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            self.set_anchorOf(o)
        self.set_uri(uri)
        self.set_referenceContext(referenceContext)

        return self


# class NifTitle(NifStructure):

#     def __init__(self, uri: str=None):
#         super().__init__(uri)


class NifWord(NifStructure):
    """
    A NIF Word

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param beginIndex: the start index in the context string

    :param endIndex: the end index in the context string

    :param referenceContext: the context to which the string refers

    :param nifsentence: the sentence of the word

    :param lemma: the lemma of the word

    :param pos: the part-of-speech tags (a list)

    :param morphofeats: the morphological features (a list)

    :param dependency: dependency relations of the word (a list)

    :param dependencyRelationType: the type of dependency relation of the word

    :param nextWord: the next word in the sentence

    :param previousWord: the previous word in the sentence

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        beginIndex: Union[Literal, int] = None,
        endIndex: Union[Literal, int] = None,
        referenceContext: NifContext = None,
        nifsentence: NifSentence = None,
        lemma: str = None,
        pos: list = None,
        morphofeats: list = None,
        dependency: list = None,
        dependencyRelationType: str = None,
        nextWord: str = None,
        previousWord: str = None,
    ):
        self.set_nifsentence(nifsentence)
        self.set_lemma(lemma)
        self.set_pos(pos)
        self.set_morphofeats(morphofeats)
        self.set_dependency(dependency)
        self.set_dependencyRelationType(dependencyRelationType)
        self.set_nextWord(nextWord)
        self.set_previousWord(previousWord)
        super().__init__(
            URIScheme=URIScheme,
            uri=uri,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=referenceContext,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:Word) uri = {self.uri}\n"
        if self.referenceContext is not None:
            s += f"  referenceContext : {self.referenceContext.uri}\n"
        if self.nifsentence is not None:
            s += f"  nifsentence : {self.nifsentence.uri}\n"
        if self.beginIndex is not None:
            s += f"  beginIndex : {self.beginIndex}\n"
        if self.endIndex is not None:
            s += f"  endIndex : {self.endIndex}\n"
        if self.nextWord is not None:
            s += f'  nextWord : "{self.nextWord.anchorOf}"\n'
        if self.previousWord is not None:
            s += f'  previousWord : "{self.previousWord.anchorOf}"\n'
        if self.anchorOf is not None:
            s += f'  anchorOf : "{self.anchorOf}"\n'
        if self.lemma is not None:
            s += f'  lemma : "{self.lemma}"\n'
        if self.pos is not None and self.pos != []:
            pos = ", ".join([str(m).replace(OLIA, "olia:") for m in self.pos])
            s += f'  pos : {pos}\n'
        if self.morphofeats is not None and self.morphofeats != []:
            mf = ", ".join([str(m).replace(OLIA, "olia:") for m in self.morphofeats])
            s += f'  morphofeats : {mf}\n'
        if self.dependency is not None and self.dependency != []:
            dep = ", ".join([dep.uri for dep in self.dependency])
            s += f'  dependency : {dep}\n'
        if self.dependencyRelationType is not None:
            s += f"  dependencyRelationtype : {self.dependencyRelationType}\n"
        return s

    @property
    def nifsentence(self):
        """
        Returns the sentence to which the word belongs
        """
        if self._nifsentence is not None:
            return self._nifsentence
        else:
            return None

    @property
    def lemma(self):
        """
        Returns the lemma of the word
        """
        if self._lemma is not None:
            return self._lemma.value
        else:
            return None

    @property
    def pos(self):
        """
        Returns the part-of-speech (pos) of the word
        """
        if self._pos is not None:
            return self._pos
        else:
            return None

    @property
    def morphofeats(self):
        """
        Returns the morphological features of the word as a list
        """
        if self._morphofeats is not None:
            return self._morphofeats
        else:
            return []

    @property
    def dependency(self):
        """
        Returns the dependencies of the word as a list
        """
        if self._dependency is not None:
            return self._dependency
        else:
            return []

    @property
    def dependencyRelationType(self):
        """
        Returns the dependency relation type of the word
        """
        if self._dependencyRelationType is not None:
            return self._dependencyRelationType.value
        else:
            return None

    @property
    def nextWord(self):
        """
        Returns the next word of the word in the sentence
        """
        if self._nextWord is not None:
            return self._nextWord
        else:
            return None

    @property
    def previousWord(self):
        """
        Returns the previous word of the word in the sentence
        """
        if self._previousWord is not None:
            return self._previousWord
        else:
            return None

    def set_nifsentence(self, nifsentence: NifSentence = None):
        """
        Sets the sentence of which the word is a part
        """
        self._nifsentence = nifsentence

    def set_lemma(self, lemma: str = None):
        """
        Sets the lemma of the word (a string)
        """
        if lemma is not None and lemma != "":
            self._lemma = Literal(lemma, datatype=XSD.string)
        else:
            self._lemma = None

    def set_pos(self, pos: Union[URIRef, list] = None):
        """
        Sets the part-of-speech (pos) of the word
        (a rdflib.URIRef or a list of rdflib.URIRef)
        """
        if pos is not None:
            if isinstance(pos, list):
                self._pos = pos
            elif isinstance(pos, URIRef):
                if self._pos is not None:
                    self._pos.append(pos)
                else:
                    self._pos = [pos]
        else:
            self._pos = []

    def set_morphofeats(self, morphofeats: Union[URIRef, list] = None):
        """
        Sets the morphological features of the word
        (a rdflib.URIRef or a list of rdflib.URIRef)
        """
        if morphofeats is not None:
            if isinstance(morphofeats, list):
                self._morphofeats = morphofeats
            elif isinstance(morphofeats, URIRef):
                if self._morphofeats is not None:
                    self._morphofeats.append(morphofeats)
                else:
                    self._morphofeats = [morphofeats]
        else:
            self._morphofeats = []

    def set_dependency(self, dependency: list = None):
        """
        Sets the dependency of the word (a list)
        """
        if dependency is not None and dependency != []:
            self._dependency = dependency
        else:
            self._dependency = None

    def set_dependencyRelationType(self, dependencyRelationType: str = None):
        """
        Sets the dependencyRelationType of the word (a string)
        """
        if dependencyRelationType is not None and dependencyRelationType != "":
            self._dependencyRelationType = Literal(
                dependencyRelationType, datatype=XSD.string
            )
        else:
            self._dependencyRelationType = None

    def set_nextWord(self, nextWord: NifWord = None):
        """
        Sets the next word of the word in the sentence
        """
        self._nextWord = nextWord

    def set_previousWord(self, previousWord: NifWord = None):
        """
        Sets the previous word of the word in the sentence
        """
        self._previousWord = previousWord

    def add_dependency(self, dependency: URIRef = None):
        """
        Add a dependency to the list of dependencies of the word
        """
        if self.dependency is not None and self.dependency != []:
            self.dependency.append(dependency)
        else:
            self.set_dependency([dependency])

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.Word)
            if self._nifsentence is not None:
                yield (self.uri, NIF.sentence, self._nifsentence.uri)
            super().triples()
            if self._anchorOf is not None:
                yield (self.uri, NIF.anchorOf, self._anchorOf)
            if self.anchorOf_no_accents is not None:
                yield (self.uri, NIF.anchorOf_no_accents, self.anchorOf_no_accents)
            if self.anchorOf_no_diacritics is not None:
                yield (
                    self.uri,
                    NIF.anchorOf_no_diacritics,
                    self.anchorOf_no_diacritics,
                )
            if self._lemma is not None:
                yield (self.uri, NIF.lemma, self._lemma)
            if self._pos is not None and self._pos != []:
                for pos in self._pos:
                    yield (self.uri, NIF.pos, pos)
            if self._morphofeats is not None and self._morphofeats != []:
                for morphofeat in self._morphofeats:
                    yield (self.uri, NIF.oliaLink, morphofeat)
            if self.nextWord is not None:
                yield (self.uri, NIF.nextWord, self.nextWord.uri)
            if self.previousWord is not None:
                yield (self.uri, NIF.previousWord, self.previousWord.uri)
            if self._dependencyRelationType is not None:
                yield (
                    self.uri,
                    NIF.dependencyRelationType,
                    self._dependencyRelationType,
                )
            if self._dependency is not None:
                for dep in self._dependency:
                    yield (self.uri, NIF.dependency, dep.uri)

    def load(
        self,
        graph: Graph = None,
        referenceContext: NifContext = None,
        nifsentence: NifSentence = None,
        uri: URIRef = None,
    ):
        """
        Load a word from a graph (based on uri of sentence)
        """
        for s, p, o in graph.triples((uri, NIF.beginIndex, None)):
            self.set_beginIndex(o)
        for s, p, o in graph.triples((uri, NIF.endIndex, None)):
            self.set_endIndex(o)
        for s, p, o in graph.triples((uri, NIF.anchorOf, None)):
            anchorOf = o
        for s, p, o in graph.triples((uri, NIF.nextWord, None)):
            self.set_nextWord(o)
        for s, p, o in graph.triples((uri, NIF.previousWord, None)):
            self.set_previousWord(o)
        for s, p, o in graph.triples((uri, NIF.lemma, None)):
            self.set_lemma(o)
        for s, p, o in graph.triples((uri, NIF.pos, None)):
            self.set_pos(o)
        for s, p, o in graph.triples((uri, NIF.oliaLink, None)):
            self.set_morphofeats(o)
        for s, p, o in graph.triples((uri, NIF.dependency, None)):
            self.add_dependency(o)
        for s, p, o in graph.triples((uri, NIF.dependencyRelationType, None)):
            self.set_dependencyRelationType(o)
        self.set_uri(uri)
        self.set_referenceContext(referenceContext)
        self.set_nifsentence(nifsentence)

        if self.anchorOf != str(anchorOf):
            logging.error(
                "Error: anchorOf in rdf data ("
                + self.anchorOf
                + ") not anchorOf of in string ("
                + anchorOf
                + ")."
            )

        return self


class NifContextCollection(NifBase):
    """
    A NIF Context Collection

    :param URIScheme: the URIScheme of the object

    :param uri: the uri of the object

    :param hasContext: the list of contexts of the collection

    :param conformsTo: the NIF Ontology version

    """

    def __init__(
        self,
        URIScheme: str = None,
        uri: Union[URIRef, str] = None,
        hasContext: list = None,
        conformsTo: Union[URIRef, str] = None,
    ):
        self.set_hasContext(hasContext)
        self.set_conformsTo(conformsTo)
        super().__init__(uri=uri)

    @property
    def hasContext(self):
        if self._hasContext is not None:
            return list(self._hasContext.values())
        else:
            return []

    def set_hasContext(self, hasContext: list = None):
        if hasContext is not None:
            self._hasContext = {context.uri: context for context in hasContext}
        else:
            self._hasContext = {}

    @property
    def conformsTo(self):
        if self._conformsTo is not None:
            return self._conformsTo
        else:
            return None

    def set_conformsTo(self, conformsTo: Union[URIRef, str]):
        if conformsTo is not None:
            if isinstance(conformsTo, str):
                self._conformsTo = URIRef(conformsTo)
            else:
                self._conformsTo = conformsTo
        else:
            self._conformsTo = URIRef(NIF_ONTOLOGY)

    def add_context(self, context: NifContext = None):
        if context is not None:
            self._hasContext[context.uri] = context

    @property
    def contexts(self):
        return list(self._hasContext.values())

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f"(nif:ContextCollection) uri = {self.uri}\n"
        s += f"  conformsTo : {self.conformsTo}\n"
        for context in self.hasContext[0:10]:
            s += f"  hasContext : {context.uri}\n"
        if len(self.hasContext) > 10:
            s += "  hasContext : ... \n"
        return s

    def triples(self):
        """
        Generates all the triples
        """
        if self.uri is not None:
            yield (self.uri, RDF.type, NIF.ContextCollection)
            if self.conformsTo is not None:
                yield (self.uri, DCTERMS.conformsTo, self.conformsTo)
            for context in self.hasContext:
                yield (self.uri, NIF.hasContext, context.uri)
                context.triples()



# -*- coding: utf-8 -*-

import regex
import logging
from collections import namedtuple
from io import BytesIO
from typing import Union

from lxml import etree
from pdfminer.converter import XMLConverter
from pdfminer.layout import LAParams
from pdfminer.pdfinterp import PDFPageInterpreter, PDFResourceManager
from pdfminer.pdfpage import PDFPage


class PDFDocument:
    def __init__(
        self,
        file: Union[str, BytesIO] = None,
        codec: str = "utf-8",
        password: str = "",
        laparams: LAParams = LAParams(),
        join_hyphenated_words=True,
    ):

        """Function to convert pdf to xml or text

        Args:

            file: location or stream of the file to be converted
            codec: codec to be used to conversion
            password: password to be used for conversion
            laparams: laparams for the pdfminer.six parser
            join_hyphenated_words: Join 'hyhen-\\n ated wor- \\nds' to 'hyphenated words'

        Returns:

        """
        self.join_hyphenated_words = join_hyphenated_words

        self.PDF_offset = namedtuple("PDF_offset", ["beginIndex", "endIndex"])

        rsrcmgr = PDFResourceManager()
        retstr = BytesIO()
        device = XMLConverter(rsrcmgr, retstr, codec=codec, laparams=laparams)

        if isinstance(file, str):
            fp = open(file, "rb")
        else:
            fp = BytesIO(file)

        interpreter = PDFPageInterpreter(rsrcmgr, device)
        maxpages = 0
        caching = True
        pagenos = set()
        for page in PDFPage.get_pages(
            fp,
            pagenos,
            maxpages=maxpages,
            password=password,
            caching=caching,
            check_extractable=False,
        ):
            interpreter.process_page(page)

        # in case the file is opened, it is closed (a stream is not closed)
        if not isinstance(file, BytesIO):
            fp.close()
        device.close()

        result = retstr.getvalue()
        retstr.close()

        self.tree = etree.fromstring(result)

    @property
    def text(self):

        # setup regexes
        CONTROL = regex.compile("[\x00-\x08\x0b-\x0c\x0e-\x1f]")
        _hyphens = "\u00AD\u058A\u05BE\u0F0C\u1400\u1806\u2010\u2011\u2012\u2e17\u30A0-"
        _hyphen_newline = regex.compile(
            r"(?<=\p{L})[" + _hyphens + "][ \t\u00a0\r]*\n{1,2}[ \t\u00a0]*(?=\\p{L})"
        )

        text = []
        for page in self.tree:
            for textbox in page:
                if textbox.tag == "textbox":
                    for textline in textbox:
                        for text_element in textline:
                            text.append(text_element.text)
                    text.append("\n")
                elif textbox.tag == "figure":
                    for text_element in textbox:
                        text.append(text_element.text)
                elif textbox.tag == "textline":
                    for text_element in textbox:
                        text.append(text_element.text)
        text = "".join([t for t in text if t is not None])

        # delete control characters
        text = CONTROL.sub("", text)

        # delete hyphens
        if self.join_hyphenated_words:
            text = _hyphen_newline.subn("", text)[0]

        return text

    @property
    def page_offsets(self):

        # setup regexes
        CONTROL = regex.compile("[\x00-\x08\x0b-\x0c\x0e-\x1f]")
        _hyphens = "\u00AD\u058A\u05BE\u0F0C\u1400\u1806\u2010\u2011\u2012\u2e17\u30A0-"
        _hyphen_newline = regex.compile(
            r"(?<=\p{L})[" + _hyphens + "][ \t\u00a0\r]*\n{1,2}[ \t\u00a0]*(?=\\p{L})"
        )

        page_offsets = []
        text = ""
        page_start_correction = 0
        page_end_correction = 0
        for page in self.tree:
            page_start = len(text)
            for textbox in page:
                if textbox.tag == "textbox":
                    for textline in textbox:
                        for text_element in textline:
                            if text_element.text is not None:
                                text += CONTROL.sub("", text_element.text)
                    text += "\n"
                elif textbox.tag == "figure":
                    for text_element in textbox:
                        if text_element.text is not None:
                            text += CONTROL.sub("", text_element.text)
                elif textbox.tag == "textline":
                    for text_element in textbox:
                        if text_element.text is not None:
                            text += CONTROL.sub("", text_element.text)
            page_end = len(text)

            if self.join_hyphenated_words:
                # retrieve all hyphens in text and calculate correction
                text_hyphens = regex.finditer(_hyphen_newline, text)
                page_end_correction = sum(
                    [hyphen.end() - hyphen.start() for hyphen in text_hyphens]
                )
                if logging.DEBUG and page_end_correction > 0:
                    logging.debug(
                        "nifigator.pdfparser.page_offsets: page_start "
                        + str(page_start)
                        + " corrected with "
                        + str(page_start_correction)
                    )
                    logging.debug(
                        "nifigator.pdfparser.page_offsets: page_end   "
                        + str(page_end)
                        + " corrected with "
                        + str(page_end_correction)
                    )
                # append corrected page offsets
                page_offsets.append(
                    self.PDF_offset(
                        page_start - page_start_correction,
                        page_end - page_end_correction,
                    )
                )
                # set page_start_correction for next page
                page_start_correction = page_end_correction
            else:
                # append page offsets
                page_offsets.append(self.PDF_offset(page_start, page_end))

        return page_offsets

    # @property
    # def paragraph_offsets(self):
    #     paragraph_offsets = []
    #     text = ""
    #     for page in self.tree:
    #         paragraph_start = len(text)
    #         for textbox in page:
    #             if textbox.tag == "textbox":
    #                 for textline in textbox:
    #                     for text_element in textline:
    #                         if text_element.text is not None:
    #                             text += self.CONTROL.sub("", text_element.text)
    #                     if (len(textline[-2].text.strip()) > 0) and (
    #                         textline[-2].text.strip()[-1] in [".", "?"]
    #                     ):
    #                         paragraph_end = len(text)
    #                         paragraph_offsets.append(
    #                             self.PDF_offset(paragraph_start, paragraph_end)
    #                         )
    #                         paragraph_start = len(text)
    #                 text += "\n"
    #             elif textbox.tag == "figure":
    #                 for text_element in textbox:
    #                     if text_element.text is not None:
    #                         text += self.CONTROL.sub("", text_element.text)
    #             elif textbox.tag == "textline":
    #                 for text_element in textbox:
    #                     if text_element.text is not None:
    #                         text += self.CONTROL.sub("", text_element.text)

    #     return paragraph_offsets



# coding: utf-8

"""converters"""

import logging
from rdflib.namespace import DC, DCTERMS, XSD
from rdflib.term import Literal, URIRef

from .const import OLIA, mapobject
from .nafdocument import NafDocument
from .nifobjects import (
    NifContext,
    NifContextCollection,
    NifPage,
    NifParagraph,
    NifPhrase,
    NifSentence,
    NifWord,
)


def nafConverter(
    collection_name: str = None,
    context_name: str = None,
    nafdocument: NafDocument = None,
    base_uri: str = None,
    base_prefix: str = None,
    URIScheme: str = None,
):

    collection_uri = base_uri + collection_name
    context_uri = base_uri + context_name

    # create nif:collection
    nif_collection = NifContextCollection(uri=collection_uri)

    # create nif:context
    if nafdocument.raw is None:
        doc_raw = ""
    else:
        doc_raw = nafdocument.raw

    # create NifContext
    nif_context = NifContext(
        isString=doc_raw,
        uri=URIRef(context_uri),
        URIScheme=URIScheme,
    )
    nif_context.set_referenceContext(nif_context)
    nif_collection.add_context(nif_context)

    # set metadata
    metadata = nafdocument.header["public"]
    metadata = {
        URIRef(key.replace("{", "").replace("}", "")): Literal(
            metadata[key], datatype=XSD.string
        )
        for key in metadata.keys()
    }
    metadata[DC.language] = Literal(nafdocument.language, datatype=XSD.string)
    metadata[DCTERMS.created] = Literal(
        nafdocument.header["fileDesc"]["creationtime"], datatype=XSD.string
    )
    metadata[DCTERMS.provenance] = Literal(
        nafdocument.header["fileDesc"]["filename"], datatype=XSD.string
    )
    # correction in naf files, dc:uri is incorrect, should be dcterms:URI
    metadata[DCTERMS.URI] = metadata[URIRef("http://purl.org/dc/elements/1.1/uri")]
    del metadata[URIRef("http://purl.org/dc/elements/1.1/uri")]
    metadata[DCTERMS.identifier] = Literal(nif_context.uri, datatype=XSD.string)
    nif_context.set_metadata(metadata)

    # create nif:sentence and nif:word
    doc_words = {word["id"]: word for word in nafdocument.text}
    doc_terms = {term["id"]: term for term in nafdocument.terms}

    doc_sentences = nafdocument.sentences
    for sent_idx, sentence in enumerate(doc_sentences):
        beginIndex = int(doc_words[sentence["span"][0]["id"]]["offset"])
        endIndex = int(doc_words[sentence["span"][-1]["id"]]["offset"]) + int(
            doc_words[sentence["span"][-1]["id"]]["length"]
        )
        nif_sentence = NifSentence(
            uri=base_uri + context_name,
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=nif_context,
            URIScheme=URIScheme
            # annotation reference missing
        )
        sentence["nif"] = nif_sentence

    nif_words = dict()
    for sent_idx, sentence in enumerate(doc_sentences):

        nif_context.add_sentence(sentence["nif"])

        # Add nextSentence and previousSentence to make graph traversable
        if sent_idx < len(doc_sentences) - 1:
            sentence["nif"].set_nextSentence(doc_sentences[sent_idx + 1]["nif"])
        if sent_idx > 0:
            sentence["nif"].set_previousSentence(doc_sentences[sent_idx - 1]["nif"])

        for word_idx, word_id in enumerate(sentence["span"]):
            word = doc_words[word_id["id"]]
            beginIndex = int(word["offset"])
            endIndex = int(word["offset"]) + int(word["length"])
            nif_word = NifWord(
                beginIndex=beginIndex,
                endIndex=endIndex,
                referenceContext=nif_context,
                nifsentence=sentence["nif"],
                # annotation reference missing
                uri=base_uri + context_name,
                URIScheme=URIScheme,
            )
            word["nif"] = nif_word
            nif_words[nif_word.uri] = nif_word

        # Add nextWord and previousWord
        for word_idx, word_id in enumerate(sentence["span"]):
            word = doc_words[word_id["id"]]
            sentence["nif"].add_word(word["nif"])

            if word_idx < len(sentence["span"]) - 1:
                word["nif"].set_nextWord(
                    doc_words[sentence["span"][word_idx + 1]["id"]]["nif"]
                )
            if word_idx > 0:
                word["nif"].set_previousWord(
                    doc_words[sentence["span"][word_idx - 1]["id"]]["nif"]
                )

        for term in sentence["terms"]:
            term_words = [s["id"] for s in doc_terms[term["id"]]["span"]]
            beginIndex = int(doc_words[term_words[0]]["offset"])
            endIndex = int(doc_words[term_words[-1]]["offset"]) + int(
                doc_words[term_words[-1]]["length"]
            )
            term_lemma = doc_terms[term["id"]].get("lemma", None)
            term_pos = doc_terms[term["id"]].get("pos", None)
            term_pos = mapobject("pos", term_pos.lower()).replace("olia:", "")
            term_pos = [OLIA[term_pos]]
            term_morphofeats = []
            morphofeats = doc_terms[term["id"]].get("morphofeat", None)
            if morphofeats is not None:
                for feat in morphofeats.split("|"):

                    if (
                        feat.split("=")[0] in ["Foreign", "Reflex", "Poss", "Abbr"]
                        and feat.split("=")[1] == "Yes"
                    ):
                        olia_term = (
                            feat.split("=")[0]
                            .replace("Poss", "PossessivePronoun")
                            .replace("Abbr", "Abbreviation")
                            .replace("Reflex", "ReflexivePronoun")
                        )
                        term_morphofeats.append(olia_term)
                    else:
                        term_morphofeats.append(
                            mapobject(feat.split("=")[0], feat.split("=")[1]).replace(
                                "olia:", ""
                            )
                        )

            term_morphofeats = [OLIA[m] for m in term_morphofeats]
            nif_term = NifWord(
                beginIndex=beginIndex,
                endIndex=endIndex,
                referenceContext=nif_context,
                lemma=term_lemma,
                pos=term_pos,
                morphofeats=term_morphofeats,
                # annotation reference missing
                uri=base_uri + context_name,
                URIScheme=URIScheme,
            )
            doc_terms[term["id"]]["nif"] = nif_term

            if nif_term.uri not in nif_words.keys():
                nif_words[nif_term.uri] = nif_term
            else:
                nif_words[nif_term.uri].set_lemma(term_lemma)
                nif_words[nif_term.uri].set_pos(term_pos)
                nif_words[nif_term.uri].set_morphofeats(term_morphofeats)

    # create nif:page
    nif_pages = []
    if len(nafdocument.text) > 0:
        page_number = int(nafdocument.text[0]["page"])
        page_start = int(nafdocument.text[0]["offset"])
        page_end = int(nafdocument.text[0]["offset"])
    else:
        page_number = 1
        page_start = 0
        page_end = 0
    beginIndex = page_start
    endIndex = page_end
    for word in nafdocument.text:
        if int(word["page"]) != page_number:
            nif_page = NifPage(
                beginIndex=beginIndex,
                endIndex=endIndex,
                referenceContext=nif_context,
                uri=base_uri + context_name,
                URIScheme=URIScheme,
            )
            beginIndex = int(word["offset"])
            endIndex = int(word["offset"]) + int(word["length"])
            nif_pages.append(nif_page)
            page_number += 1
        endIndex = int(word["offset"]) + int(word["length"])
    nif_page = NifPage(
        beginIndex=beginIndex,
        endIndex=endIndex,
        referenceContext=nif_context,
        uri=base_uri + context_name,
        URIScheme=URIScheme,
    )
    nif_pages.append(nif_page)

    nif_context.set_Pages(page for page in nif_pages)

    # create nif:phrases
    nif_phrases = []
    for entity in nafdocument.entities:
        taClassRef = "https://stanfordnlp.github.io/stanza#" + entity.get(
            "type", "unknown"
        )
        entity_words = [
            ss["id"] for s in entity["span"] for ss in doc_terms[s["id"]]["span"]
        ]
        beginIndex = int(doc_words[entity_words[0]]["offset"])
        endIndex = int(doc_words[entity_words[-1]]["offset"]) + int(
            doc_words[entity_words[-1]]["length"]
        )
        nif_phrase = NifPhrase(
            beginIndex=beginIndex,
            endIndex=endIndex,
            referenceContext=nif_context,
            taClassRef=URIRef(taClassRef),
            entityOccurrence=True,
            uri=base_uri + context_name,
            URIScheme=URIScheme,
        )
        nif_phrases.append(nif_phrase)
    nif_context.set_Phrases(nif_phrases)

    # Add dependencies:
    for dep in nafdocument.deps:
        from_term = doc_terms[dep["from_term"]]
        to_term = doc_terms[dep["to_term"]]
        rfunc = dep["rfunc"]
        if "nif" in from_term.keys() and "nif" in to_term.keys():
            from_term["nif"].add_dependency(to_term["nif"])
            from_term["nif"].set_dependencyRelationType(rfunc)
        else:
            if "nif" not in from_term.keys():
                logging.warning(
                    ".. from term in dependency not found:\n" + str(from_term)
                )
            if "nif" not in to_term.keys():
                logging.warning(".. to term in dependency not found:\n" + str(to_term))

    # create nif:paragraph
    doc_paragraphs = nafdocument.paragraphs
    for para_idx, paragraph in enumerate(doc_paragraphs):

        if paragraph["span"] != []:

            beginIndex = int(doc_words[paragraph["span"][0]["id"]]["offset"])
            endIndex = int(doc_words[paragraph["span"][-1]["id"]]["offset"]) + int(
                doc_words[paragraph["span"][-1]["id"]]["length"]
            )
            nif_paragraph = NifParagraph(
                beginIndex=beginIndex,
                endIndex=endIndex,
                referenceContext=nif_context,
                # annotation reference missing
                uri=base_uri + context_name,
                URIScheme=URIScheme,
            )
            paragraph["nif"] = nif_paragraph

    nif_context.set_Paragraphs(paragraph["nif"] for paragraph in doc_paragraphs)

    return nif_collection



# coding: utf-8

"""naf document."""

import logging
import datetime

from lxml import etree

from .const import (
    ChunkElement,
    DependencyRelation,
    EntityElement,
    MultiwordElement,
    ProcessorElement,
    RawElement,
    TermElement,
    WordformElement,
)
from .utils import load_dtd, prepare_comment_text, time_in_correct_format

NAF_VERSION_TO_DTD = {
    "v3": "data/naf_v3.dtd",
    "v3.1": "data/naf_v3_1.dtd",
}

FILEDESC_ELEMENT_TAG = "fileDesc"
PUBLIC_ELEMENT_TAG = "public"

TERMS_LAYER_TAG = "terms"
TERM_OCCURRENCE_TAG = "term"

DEPS_LAYER_TAG = "deps"
DEP_OCCURRENCE_TAG = "dep"

TEXT_LAYER_TAG = "text"
TEXT_OCCURRENCE_TAG = "wf"

ENTITIES_LAYER_TAG = "entities"
ENTITY_OCCURRENCE_TAG = "entity"

CHUNKS_LAYER_TAG = "chunks"
CHUNK_OCCURRENCE_TAG = "chunk"

MULTIWORDS_LAYER_TAG = "multiwords"
MULTIWORD_OCCURRENCE_TAG = "mw"

COMPONENTS_LAYER_TAG = "components"
COMPONENT_OCCURRENCE_TAG = "component"

RAW_LAYER_TAG = "raw"
FORMATS_LAYER_TAG = "formats"
FORMATS_LAYER_COPY_TAG = "formats_copy"
NAF_HEADER = "nafHeader"

SPAN_OCCURRENCE_TAG = "span"
TARGET_OCCURRENCE_TAG = "target"
EXT_REFS_OCCURRENCE_TAG = "externalReferences"
EXT_REF_OCCURRENCE_TAG = "externalRef"

LINGUISTIC_LAYER_TAG = "linguisticProcessors"
LINGUISTIC_OCCURRENCE_TAG = "lp"

PREFIX_NAF_BASE = "naf-base"

namespaces = {
    "dc": "http://purl.org/dc/elements/1.1/",
    # "naf-base": "https://dnb.nl/naf-Base/elements/1.0/",
}


def QName(prefix: str = None, name: str = None):
    """ """
    # currently no namespaces used
    # qname = etree.QName('{'+namespaces[prefix]+'}'+name, name)
    return name


class NafDocument(etree._ElementTree):
    """The NafDocument class (subclass of an etree.elementtree)"""

    def generate(self, params: dict):
        """Initialize a NafDocument with data from the params dict"""
        self._setroot(etree.Element("NAF", nsmap=namespaces))
        self.set_version(params["naf_version"])
        if params["language"] is not None:
            self.set_language(params["language"])
        self.add_nafHeader()
        self.add_filedesc_element(params["fileDesc"])
        self.add_public_element(params["public"])

    def open(self, input: str):
        """Function to open a NafDocument

        Args:
            input: the location of the NafDocument to be opened

        Returns:
            NafDocument: the NAF document that is opened

        """
        with open(input, "r", encoding="utf-8") as f:
            self._setroot(etree.parse(f).getroot())
        return self

    def write(self, output: str) -> None:
        """Function to write a NafDocument

        Args:
            output: the location of the NafDocument to be stored

        Returns:
            None

        """
        super().write(output, encoding="utf-8", pretty_print=True, xml_declaration=True)

    @property
    def version(self):
        """Returns version of the NAF document"""
        return self.getroot().get("version")

    @property
    def language(self):
        """Returns language of the NAF document"""
        return self.getroot().get("{http://www.w3.org/XML/1998/namespace}lang")

    @property
    def header(self):
        """Returns header of the NAF document as a dict"""
        header = dict()
        ling_proc = list()
        for child in self.find(NAF_HEADER):
            if child.tag == "fileDesc":
                header["fileDesc"] = dict(child.attrib)
            if child.tag == "public":
                header["public"] = dict(child.attrib)
            if child.tag == LINGUISTIC_LAYER_TAG:
                header_data = dict(child.attrib)
                lp = list()
                for child2 in child:
                    if child2.tag == LINGUISTIC_OCCURRENCE_TAG:
                        lp.append(child2.attrib)
                header_data["lps"] = lp
                ling_proc.append(header_data)
        header[LINGUISTIC_LAYER_TAG] = ling_proc
        return header

    @property
    def raw(self):
        """Returns raw layer of the NAF document as string"""
        return self.find(RAW_LAYER_TAG).text

    @property
    def deps(self):
        """Returns dependencies layer of the NAF document as list"""
        return [
            dep.attrib
            for dep in self.findall(DEPS_LAYER_TAG + "/" + DEP_OCCURRENCE_TAG)
        ]

    @property
    def text(self):
        """Returns text layer of the NAF document as list of dicts"""
        return [
            dict({"text": wf.text}, **dict(wf.attrib))
            for wf in self.findall(TEXT_LAYER_TAG + "/" + TEXT_OCCURRENCE_TAG)
        ]

    @property
    def terms(self):
        """Returns terms layer of the NAF document as list of dicts"""
        terms = list()
        for child in self.findall(TERMS_LAYER_TAG + "/" + TERM_OCCURRENCE_TAG):
            term_data = dict(child.attrib)
            for child2 in child:
                if child2.tag == SPAN_OCCURRENCE_TAG:
                    span = [
                        child3.attrib
                        for child3 in child2
                        if child3.tag == TARGET_OCCURRENCE_TAG
                    ]
                    term_data["span"] = span
                if child2.tag == EXT_REFS_OCCURRENCE_TAG:
                    ext_refs = list()
                    for child3 in child2:
                        if child3.tag == EXT_REF_OCCURRENCE_TAG:
                            ext_refs.append(child3.attrib)
                    term_data["ext_refs"] = ext_refs
            terms.append(term_data)
        return terms

    @property
    def multiwords(self):
        """Returns multiword layer of the NAF document as list of dicts"""
        mw = list()
        for child in self.findall(
            MULTIWORDS_LAYER_TAG + "/" + MULTIWORD_OCCURRENCE_TAG
        ):
            mw_data = dict(child.attrib)
            com = list()
            for child2 in child:
                if child2.tag == COMPONENT_OCCURRENCE_TAG:
                    com_data = dict(child2.attrib)
                    for child3 in child2:
                        if child3.tag == SPAN_OCCURRENCE_TAG:
                            span = [
                                child4.attrib
                                for child4 in child3
                                if child4.tag == TARGET_OCCURRENCE_TAG
                            ]
                            com_data["span"] = span
                        if child3.tag == EXT_REFS_OCCURRENCE_TAG:
                            ext_refs = list()
                            for child4 in child3:
                                if child4.tag == EXT_REF_OCCURRENCE_TAG:
                                    ext_refs.append(child4.attrib)
                            com_data["ext_refs"] = ext_refs
                    com.append(com_data)
            mw_data["components"] = com
            mw.append(mw_data)
        return mw

    @property
    def entities(self):
        """Returns entities layer of the NAF document as list of dicts"""
        entities = list()
        for child in self.findall(ENTITIES_LAYER_TAG + "/" + ENTITY_OCCURRENCE_TAG):
            entity_data = dict(child.attrib)
            spans = list()
            ext_refs = list()
            for child2 in child:
                if child2.tag == SPAN_OCCURRENCE_TAG:
                    span = list()
                    for child3 in child2:
                        if child3.tag == etree.Comment:
                            entity_data["text"] = child3.text
                        elif child3.tag == TARGET_OCCURRENCE_TAG:
                            span.append(child3.attrib)
                    spans.append(span)
                if child2.tag == EXT_REFS_OCCURRENCE_TAG:
                    ext_ref = list()
                    for child3 in child2:
                        if child3.tag == EXT_REF_OCCURRENCE_TAG:
                            ext_ref.append(child3.attrib)
                    ext_refs.append(ext_ref)
            if ext_refs != []:
                entity_data["ext_refs"] = ext_refs
            if spans != []:
                if len(spans) == 1:
                    entity_data["span"] = spans[0]
                else:
                    entity_data["spans"] = spans
            entities.append(entity_data)
        return entities

    @property
    def sentences(self):
        """Returns sentences of the NAF document as list of dicts"""
        word2term = {
            item["id"]: term["id"] for term in self.terms for item in term["span"]
        }
        last_sentence = list()
        sentences = list()
        sentence_list = list()
        sent_num = 1
        pages = set()
        para = set()
        span = list()
        terms = list()
        for item in self.text:
            if item["sent"] == str(sent_num):
                sentence_list.append(item["text"])
                span.append({"id": item["id"]})
                if item["id"] in word2term.keys():
                    terms.append({"id": word2term.get(item["id"])})
                pages.add(item.get("page", "0"))
                para.add(item.get("para", "0"))
                last_sentence = [item["sent"]]
            else:
                sentences.append(
                    {
                        "text": " ".join(sentence_list),
                        "para": list(para),
                        "page": list(pages),
                        "span": span,
                        "terms": terms,
                        "sent": last_sentence,
                    }
                )
                sentence_list = list([item["text"]])
                pages = set()
                para = set()
                span = list()
                terms = list()
                span.append({"id": item["id"]})
                if item["id"] in word2term.keys():
                    terms.append({"id": word2term.get(item["id"])})
                pages.add(item.get("page", "0"))
                para.add(item.get("para", "0"))
                last_sentence = [item["sent"]]
                sent_num += 1
        if sent_num >= 1:
            if sentence_list != []:
                sentences.append(
                    {
                        "text": " ".join(sentence_list),
                        "para": list(para),
                        "page": list(pages),
                        "span": span,
                        "terms": terms,
                        "sent": last_sentence,
                    }
                )
        return sentences

    @property
    def paragraphs(self):
        """Returns paragraphs of the NAF document as list of dicts"""

        # return empty list if no para attributes are included
        if any(["para" not in item.keys() for item in self.text]):
            return []

        word2term = {
            item["id"]: term["id"] for term in self.terms for item in term["span"]
        }
        paragraphs = list()
        paragraph_list = list()
        sentence_list = list()
        para_num = 1
        pages = set()
        para = set()
        span = list()
        terms = list()
        for item in self.text:
            if item["para"] == str(para_num):
                paragraph_list.append(item["text"])
                span.append({"id": item["id"]})
                if item["id"] in word2term.keys():
                    terms.append({"id": word2term.get(item["id"])})
                pages.add(item.get("page", "0"))
                para.add(item.get("para", "0"))
                if item.get("sent", "0") not in sentence_list:
                    sentence_list.append(item.get("sent", "0"))
            else:
                paragraphs.append(
                    {
                        "text": " ".join(paragraph_list),
                        "para": list(para),
                        "page": list(pages),
                        "span": span,
                        "terms": terms,
                        "sent": sentence_list,
                    }
                )
                paragraph_list = list([item["text"]])
                sentence_list = list()
                pages = set()
                para = set()
                span = list()
                terms = list()
                span.append({"id": item["id"]})
                if item["id"] in word2term.keys():
                    terms.append({"id": word2term.get(item["id"])})
                pages.add(item.get("page", "0"))
                para.add(item.get("para", "0"))
                if item.get("sent", "0") not in sentence_list:
                    sentence_list.append(item.get("sent", "0"))
                para_num += 1
        if para_num > 1:
            paragraphs.append(
                {
                    "text": " ".join(paragraph_list),
                    "para": list(para),
                    "page": list(pages),
                    "span": span,
                    "terms": terms,
                    "sent": sentence_list,
                }
            )
        return paragraphs

    @property
    # @TODO: reduce complexity of code
    def formats_copy(self):
        """Returns formats_copy layer of the NAF document as list of dicts"""
        pages = list()
        for child in self.find(FORMATS_LAYER_COPY_TAG):
            # transform the page elements
            if child.tag == "page":
                pages_data = dict(child.attrib)
                textboxes = list()
                layout = list()
                # tranform the textbox elements
                for child2 in child:
                    # get textbox data
                    if child2.tag == "textbox":
                        textbox_data = dict(child2.attrib)
                        textlines = list()
                        # get textline data
                        for child3 in child2:
                            if child3.tag == "textline":
                                textline_data = dict(child3.attrib)
                                texts = list()
                                # get text data
                                for child4 in child3:
                                    if child4.tag == "text":
                                        text_data = dict(child4.attrib)
                                        text_data["text"] = child4.text
                                        texts.append(text_data)
                                textline_data["texts"] = texts
                                textlines.append(textline_data)
                        textbox_data["textlines"] = textlines
                        textboxes.append(textbox_data)

                    # transform the layout element
                    elif child2.tag == "layout":
                        layout_data = dict(child2.attrib)
                        textgroups = list()
                        # get the textgroup data
                        for child3 in child2:
                            if child3.tag == "textgroup":
                                textgroup_data = dict(child3.attrib)
                                textboxes2 = list()
                                # get the textbox data
                                for child4 in child3:
                                    if child4.tag == "textbox":
                                        textbox2_data = dict(child4.attrib)
                                        textbox2_data["textbox"] = child4.text
                                        textboxes2.append(textbox2_data)
                                textgroup_data["textboxes"] = textboxes2
                                textgroups.append(textgroup_data)
                        layout_data["textgroup"] = textgroups
                        layout.append(layout_data)

                pages_data["textboxes"] = textboxes
                pages_data["layout"] = layout
                pages.append(pages_data)

                return pages

    @property
    def formats(self):
        """Returns formats layer of the NAF document as list of dicts"""
        pages = list()
        for child in self.find(FORMATS_LAYER_TAG):
            if child.tag == "page":
                pages_data = dict(child.attrib)
                textboxes = list()
                headers = list()
                figures = list()
                tables = list()
                for child2 in child:
                    if child2.tag == "textbox":
                        textbox_data = dict(child2.attrib)
                        textlines = list()
                        for child3 in child2:
                            if child3.tag == "textline":
                                textline_data = dict(child3.attrib)
                                texts = list()
                                for child4 in child3:
                                    if child4.tag == "text":
                                        text_data = dict(child4.attrib)
                                        text_data["text"] = child4.text
                                        texts.append(text_data)
                                textline_data["texts"] = texts
                                textlines.append(textline_data)
                        textbox_data["textlines"] = textlines
                        textboxes.append(textbox_data)
                    # elif child2.tag == "layout":
                    elif child2.tag == "figure":
                        figure_data = dict(child2.attrib)
                        texts = list()
                        for child3 in child2:
                            if child3.tag in ["text", "line"]:
                                text_data = dict(child3.attrib)
                                text_data["text"] = child3.text
                                texts.append(text_data)
                        figure_data["texts"] = texts
                        figures.append(figure_data)
                    elif child2.tag == "header":
                        spans = list()
                        for child3 in child2:
                            for child4 in child3:
                                spans.append(child4.attrib)
                        headers_data = dict(child2.attrib)
                        headers_data["spans"] = spans
                        headers.append(headers_data)
                    elif child2.tag == "tables":
                        tables = list()
                        for table in child2:
                            table_data = dict(table.attrib)
                            rows = list()
                            for row in table:
                                row_data = dict(row.attrib)
                                cells = list()
                                for cell in row:
                                    cell_data = dict(cell.attrib)
                                    if cell.tag == "index":
                                        cell_data["index"] = cell.text
                                    else:
                                        cell_data["cell"] = cell.text
                                    cells.append(cell_data)
                                row_data["row"] = cells
                                rows.append(row_data)
                            table_data["table"] = rows
                            tables.append(table_data)

                pages_data["textboxes"] = textboxes
                pages_data["figures"] = figures
                pages_data["headers"] = headers
                pages_data["tables"] = tables
                pages.append(pages_data)

        return pages

    def __getattr__(self, name):
        """Return custom made layer of the NAF document"""
        layer = self.find(name)
        if layer is not None:
            layer_list = list()
            for item in layer:
                if item.text is not None:
                    item_data = dict({"text": item.text}, **dict(item.attrib))
                else:
                    item_data = dict(item.attrib)
                for span in item:
                    if span.tag == SPAN_OCCURRENCE_TAG:
                        targets = list()
                        for child3 in span:
                            if child3.tag == etree.Comment:
                                item_data["text"] = child3.text
                            elif child3.tag == TARGET_OCCURRENCE_TAG:
                                targets.append(child3.attrib)
                    item_data["span"] = targets
                layer_list.append(item_data)
            return layer_list
        return super().name

    def set_language(self, language: str):
        """Set language of the NAF document"""
        self.getroot().set("{http://www.w3.org/XML/1998/namespace}lang", language)

    def set_version(self, version):
        """Set version of the NAF document"""
        self.getroot().set("version", version)

    def tree2string(self, byte: bool = False):
        """Return xml string of the NAF document"""
        xml_string = etree.tostring(
            self, pretty_print=True, xml_declaration=True, encoding="utf-8"
        )
        if byte:
            return xml_string
        else:
            return xml_string.decode("utf-8")

    def validate(self):
        """Validate xml string of the NAF document"""
        dtd = load_dtd(NAF_VERSION_TO_DTD[self.get_version()])
        success = dtd.validate(self.getroot())
        if not success:
            logging.error("DTD error log:")
            for error in dtd.error_log.filter_from_errors():
                logging.error(str(error))
            return success
        return success

    def remove_layer_elements(self, layer: str = None):
        """Remove all elements in layer"""
        layer = self.find(layer)
        for items in layer:
            layer.remove(items)

    def get_attributes(self, data, namespace=None, exclude=list()):
        """ """
        if not isinstance(data, dict):
            data = data._asdict()
        for key, value in dict(data).items():
            if value is None:
                del data[key]
            if isinstance(value, datetime.datetime):
                data[key] = time_in_correct_format(value)
            if isinstance(value, float):
                data[key] = str(value)
            if isinstance(value, int):
                data[key] = str(value)
            if isinstance(value, list):
                del data[key]
        if namespace:
            for key, value in dict(data).items():
                qname = etree.QName("{" + namespace + "}" + key, key)
                del data[key]
                data[qname] = value
        for key in dict(data).keys():
            if key in exclude:
                del data[key]
        return data

    def layer(self, layer_tag: str):
        """ """
        layer = self.find(layer_tag)
        if layer is None:
            layer = etree.SubElement(self.getroot(), QName(PREFIX_NAF_BASE, layer_tag))
        return layer

    def subelement(
        self,
        element: etree._Element = None,
        tag: str = None,
        data={},
        attributes_to_ignore: list = [],
    ):
        """ """
        if not isinstance(data, dict):
            data = data._asdict()

        data = dict(data)
        for attr in attributes_to_ignore:
            del data[attr]

        subelement = etree.SubElement(
            element,
            QName(PREFIX_NAF_BASE, tag),
            self.get_attributes(data),
        )

        return subelement

    def add_nafHeader(self):
        """ """
        self.layer(NAF_HEADER)

    def add_filedesc_element(self, data: dict):
        """
        FILEDESC ELEMENT
            <fileDesc> is an empty element containing information about the
              computer document itself. It has the following attributes:

              - title: the title of the document (optional).
              - author: the author of the document (optional).
              - creationtime: when the document was created. In ISO 8601. (optional)
              - filename: the original file name (optional).
              - filetype: the original format (PDF, HTML, DOC, etc) (optional).
              - pages: number of pages of the original document (optional).

        ELEMENT fileDesc EMPTY

        ATTLIST fileDesc
            title CDATA #IMPLIED
            author CDATA #IMPLIED
            creationtime CDATA #IMPLIED
            filename CDATA #IMPLIED
            filetype CDATA #IMPLIED
            pages CDATA #IMPLIED
        """
        naf_header = self.find(NAF_HEADER)
        _ = self.subelement(
            element=naf_header, tag=FILEDESC_ELEMENT_TAG, data=data
        )

    def add_public_element(self, data: dict):
        """
        PUBLIC ELEMENT
            <public> is an empty element which stores public information about
            the document, such as its URI. It has the following attributes:

            - publicId: a public identifier (for instance, the number
              inserted by the capture server) (optional).
            - uri: a public URI of the document (optional).

        ELEMENT public EMPTY

        ATTLIST public
            publicId CDATA #IMPLIED
            uri CDATA #IMPLIED

        Difference to NAF: here all attributes are mapped to the Dublic Core
        """
        self.subelement(
            element=self.find(NAF_HEADER),
            tag=PUBLIC_ELEMENT_TAG,
            data=self.get_attributes(data, "http://purl.org/dc/elements/1.1/"),
        )

    def add_processor_element(self, layer: str, data: ProcessorElement):
        """
        LINGUISTICPROCESSORS ELEMENT
            <linguisticProcessors> elements store the information
            about which linguistic processors
            produced the NAF document. There can be several
            <linguisticProcessors> elements, one
              per NAF layer. NAF layers correspond to the top-level
              elements of the documents, such as "text", "terms", "deps" etc.

        ELEMENT linguisticProcessors (lp)+

        ATTLIST linguisticProcessors
            layer CDATA #REQUIRED

        LP ELEMENT
             <lp> elements describe one specific linguistic processor. <lp> elements
                 have the following attributes:

                 - name: the name of the processor
                 - version: processor's version
                 - timestamp: a timestamp, denoting the date/time at
                 which the processor was launched. The timestamp
                 follows the XML Schema xs:dateTime type (See
                 http://www.w3.org/TR/xmlschema-2/#isoformats). In summary, the
                 date is specified following the form "YYYY-MM-DDThh:mm:ss" (all
                 fields required). To specify a time zone, you can either enter
                 a dateTime in UTC time by adding a "Z" behind the time
                 ("2002-05-30T09:00:00Z") or you can specify an offset from
                 the UTC time by adding a positive or negative time
                 behind the time ("2002-05-30T09:00:00+06:00").
                 - beginTimestamp (optional): a timestamp, denoting the date/time at
                 which the processor started the process. It follows the XML Schema
                 xs:dateTime format.
                 - endTimestamp (optional): a timestamp, denoting the date/time at
                 which the processor ended the process. It follows the XML Schema
                 xs:dateTime format.

        ELEMENT lp EMPTY

        ATTLIST lp
            name CDATA #REQUIRED
            version CDATA #REQUIRED
            timestamp CDATA #IMPLIED
            beginTimestamp CDATA #IMPLIED
            endTimestamp CDATA #IMPLIED
            hostname CDATA #IMPLIED
        """
        proc = self.subelement(
            element=self.find(NAF_HEADER),
            tag=LINGUISTIC_LAYER_TAG,
            data={"layer": layer},
        )
        _ = self.subelement(element=proc, tag=LINGUISTIC_OCCURRENCE_TAG, data=data)

    def add_wf_element(self, data: WordformElement, cdata: bool):
        """
        WORDFORM ELEMENT
            <wf> elements describe and contain all word foorms generated after
            the tokenization step
              <wf> elements have the following attributes:
                - id: the id of the word form (REQUIRED and UNIQUE)
                - sent: sentence id of the word form (optional)
                - para: paragraph id of the word form (optional)
                - page: page id of the word form (optional)
                - offset: the offset (in characters) of the word form (optional)
                - length: the length (in characters) of the word form (optional)
                - xpath: in case of source xml files, the xpath expression
                identifying the original word form (optional)

        ELEMENT wf (#PCDATA|subtoken)*

        ATTLIST wf
            id ID #REQUIRED
            sent CDATA #IMPLIED
            para CDATA #IMPLIED
            page CDATA #IMPLIED
            offset CDATA #REQUIRED
            length CDATA #REQUIRED
            xpath CDATA #IMPLIED
        """
        wf = self.subelement(
            element=self.layer(TEXT_LAYER_TAG),
            tag=TEXT_OCCURRENCE_TAG,
            data=data,
            attributes_to_ignore=["text"],
        )

        wf.text = (
            etree.CDATA(data.text if "]]>" not in data.text else " " * len(data.text))
            if cdata
            else data.text
        )

    def add_raw_text_element(self, data: RawElement):
        """
        RAW ELEMENT

        ELEMENT raw (#PCDATA)
        """
        self.layer(RAW_LAYER_TAG).text = data.text

    def add_dependency_element(self, data: DependencyRelation, comments: bool):
        """
        DEPS ELEMENT

        ELEMENT deps (dep)+

        DEP ELEMENT
            The <dep> elements have the following attributes:
            -   from: term id of the source element (REQUIRED)
            -   to: term id of the target element (REQUIRED)
            -   rfunc: relational function.(REQUIRED)
            -       case: declension case (optional)

        ELEMENT dep EMPTY
        ATTLIST dep
            from IDREF #REQUIRED
            to IDREF #REQUIRED
            rfunc CDATA #REQUIRED
            case CDATA #IMPLIED
        """
        layer = self.layer(DEPS_LAYER_TAG)

        if comments:
            layer.append(etree.Comment(data.comment))

        _ = self.subelement(
            element=layer,
            tag=DEP_OCCURRENCE_TAG,
            data=data,
            attributes_to_ignore=["comment"],
        )

    def add_entity_element(self, data: EntityElement, naf_version: str, comments: str):
        """
        ENTITY ELEMENT
            A named entity element has the following attributes:
            -   id: the id for the named entity (REQUIRED)
            -   type:  type of the named entity. (IMPLIED) Currently,
            8 values are possible:
            -   Person
            -   Organization
            -   Location
            -   Date
            -   Time
            -   Money
            -   Percent
            -   Misc

        ELEMENT entity (span|externalReferences)+

        ATTLIST entity
            id ID #REQUIRED
            type CDATA #IMPLIED
            status CDATA #IMPLIED
            source CDATA #IMPLIED
        """
        element = self.subelement(
            element=self.layer(ENTITIES_LAYER_TAG), tag=ENTITY_OCCURRENCE_TAG, data=data
        )

        if data.span != []:
            self.add_span_element(
                element=element, data=data, comments=comments, naf_version=naf_version
            )

        if data.ext_refs != []:
            self.add_external_reference_element(element=element, ext_refs=data.ext_refs)

    def add_term_element(
        self, data: TermElement, layer_to_attributes_to_ignore: dict, comments: bool
    ):
        """
        TERM ELEMENT
            attributes of term elements
            id: unique identifier (REQUIRED AND UNIQUE)
            type: type of the term. (IMPLIED) Currently, 3 values are possible:
            open: open category term
            close: close category term
            lemma: lemma of the term (IMPLIED).
            pos: part of speech. (IMPLIED)
            Users are encourage to provide URIs to part of speech values to
            dereference these them.
            more complex pos attributes may be formed by concatenating values
            separated by a dot ".".
            morphofeat: morphosyntactic feature encoded as a single attribute.
            case: declension case of the term (optional).
            head: if the term is a compound, the id of the head component
            (optional).
            component_of: if the term is part of multiword, i.e., referenced
            by a multiwords/mw element
            than this attribute can be used to make reference to the multiword.
            compound_type: endocentric or exocentric

        ELEMENT term (sentiment?|span|externalReferences|component)+
        ATTLIST term
            id ID #REQUIRED
            type CDATA #IMPLIED
            lemma CDATA #IMPLIED
            pos CDATA #IMPLIED
            morphofeat CDATA #IMPLIED
            netype CDATA #IMPLIED
            case CDATA #IMPLIED
            head CDATA #IMPLIED
            component_of IDREF #IMPLIED
            compound_type CDATA #IMPLIED
        """
        element = self.subelement(
            element=self.layer(TERMS_LAYER_TAG),
            tag=TERM_OCCURRENCE_TAG,
            data=data,
            attributes_to_ignore=layer_to_attributes_to_ignore.get("terms", list()),
        )

        if data.span != []:
            self.add_span_element(element=element, data=data, comments=comments)

        if data.ext_refs != []:
            self.add_external_reference_element(element=element, ext_refs=data.ext_refs)

    def add_chunk_element(self, data: ChunkElement, comments: bool):
        """
        CHUNKS ELEMENT

        ELEMENT chunks (chunk)+

        CHUNK ELEMENT
            The <chunk> elements have the following attributes:
            -   id: unique identifier (REQUIRED)
            -   head: the chunk head’s term id  (REQUIRED)
            -   phrase: type of the phrase (REQUIRED)
            -   case: declension case (optional)

        ELEMENT chunk (span)+

        ATTLIST chunk
            id ID #REQUIRED
            head IDREF #REQUIRED
            phrase CDATA #REQUIRED
            case CDATA #IMPLIED
        """
        element = self.subelement(
            element=self.layer(CHUNKS_LAYER_TAG), tag=CHUNK_OCCURRENCE_TAG, data=data
        )

        if data.span != []:
            self.add_span_element(element=element, data=data, comments=comments)

    def add_span_element(self, element, data, comments=False, naf_version: str = None):
        """
        SPAN ELEMENT

        ELEMENT span (target)+

        ATTLIST span
            primary CDATA #IMPLIED
            status CDATA #IMPLIED
        """
        if not isinstance(data, dict):
            data = data._asdict()

        if (naf_version is not None) and (naf_version == "v3"):
            references = self.subelement(element=element, tag="references")
            span = self.subelement(element=references, tag=SPAN_OCCURRENCE_TAG)
        else:
            span = self.subelement(element=element, tag=SPAN_OCCURRENCE_TAG)
        if comments:
            comment = " ".join(data["comment"])
            comment = prepare_comment_text(comment)
            span.append(etree.Comment(comment))

        for target in data["span"]:
            self.subelement(
                element=span, tag=TARGET_OCCURRENCE_TAG, data={"id": target}
            )

    def add_external_reference_element(self, element, ext_refs: list):
        """
        EXTERNALREFERENCES ELEMENT
            The <externalReferences> element is used to associate terms to
            external resources, such as elements of a Knowledge base, an ontology,
            etc. It consists of several <externalRef> elements, one per
            association.

        ELEMENT externalReferences (externalRef)+

        EXTERNALREF ELEMENT
            <externalRef> elements have the following attributes:
            - resource: indicates the identifier of the resource
            referred to.
            - reference: code of the referred element. If the element is a
            synset of some version of WordNet, it follows the pattern:
            [a-z]{3}-[0-9]{2}-[0-9]+-[nvars]
            which is a string composed by four fields separated by a dash.
            The four fields are the following:
            - Language code (three characters).
            - WordNet version (two digits).
            - Synset identifier composed by digits.
            - POS character:
            n noun
            v verb
            a adjective
            r adverb
            examples of valid patterns are: ``ENG-20-12345678-n'',
            ``SPA-16-017403-v'', etc.
            - confidence: a floating number between 0 and 1.
              Indicates the confidence weight of the association

        ELEMENT externalRef (sentiment|externalRef)*

        ATTLIST externalRef
            resource CDATA #IMPLIED
            reference CDATA #REQUIRED
            reftype CDATA #IMPLIED
            status CDATA #IMPLIED
            source CDATA #IMPLIED
            confidence CDATA #IMPLIED
            timestamp CDATA #IMPLIED
        """
        if not isinstance(ext_refs, list):
            logging.info("ext_refs should be a list of dictionaries (can be empty)")

        ext_refs_el = self.subelement(element=element, tag="externalReferences")
        for ext_ref in ext_refs:
            ext_ref_el = self.subelement(
                element=ext_refs_el,
                tag="externalRef",
                data={"reference": ext_ref["reference"]},
            )
            for optional_attr in ["resource", "source", "timestamp"]:
                if optional_attr in ext_ref:
                    ext_ref_el.set(optional_attr, ext_ref[optional_attr])

    def add_multiword_element(self, data: MultiwordElement):
        """
        MULTIWORDS ELEMENT

        ELEMENT multiwords (mw)+

        MW ELEMENT
            attributes of mw elements
            id: unique identifier (REQUIRED AND UNIQUE)
            lemma: lemma of the term (IMPLIED).
            pos: part of speech. (IMPLIED)
            morphofeat: morphosyntactic feature encoded as a single attribute. (IMPLIED)
            case: declension case (IMPLIED)
            status: manual | system | deprecated
            type: phrasal, idiom

        ELEMENT mw (component|externalReferences)+

        ATTLIST mw
            id ID #REQUIRED
            lemma CDATA #IMPLIED
            pos CDATA #IMPLIED
            morphofeat CDATA #IMPLIED
            case CDATA #IMPLIED
            status CDATA #IMPLIED
            type CDATA #REQUIRED
        """
        mw = self.subelement(
            element=self.layer(MULTIWORDS_LAYER_TAG),
            tag=MULTIWORD_OCCURRENCE_TAG,
            data=data,
        )
        for component in data.components:
            com = self.subelement(
                element=mw, tag=COMPONENT_OCCURRENCE_TAG, data=component
            )
            self.add_span_element(element=com, data=component)



import pytest

import nifigator

from rdflib.term import URIRef
import stanza

def test_nif_context():
    """API Tests"""
    context = nifigator.NifContext(
        uri="https://mangosaurus.eu/rdf-data/doc_1",
        URIScheme=nifigator.OffsetBasedString,
        isString="The cat sat on the mat. Felix was his name.",
    )
    assert type(context) == nifigator.NifContext
    assert context.isString == "The cat sat on the mat. Felix was his name."
    assert context.uri == URIRef("https://mangosaurus.eu/rdf-data/doc_1")
    assert context.URIScheme == nifigator.OffsetBasedString

def test_nif_collection():

    context = nifigator.NifContext(
        uri="https://mangosaurus.eu/rdf-data/doc_1",
        URIScheme=nifigator.OffsetBasedString,
        isString="The cat sat on the mat. Felix was his name.",
    )
    collection = nifigator.NifContextCollection(uri="https://mangosaurus.eu/rdf-data")
    collection.add_context(context)
    assert type(collection) == nifigator.NifContextCollection
    assert collection.conformsTo == URIRef('http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core/2.1')
    assert collection.contexts == [context]

def test_load_from_dict():

    context = nifigator.NifContext(
        uri="https://mangosaurus.eu/rdf-data/doc_1",
        URIScheme=nifigator.OffsetBasedString,
        isString="The cat sat on the mat. Felix was his name.",
    )

    nlp = stanza.Pipeline("en", verbose=False)
    stanza_dict = nlp(context.isString).to_dict()
    context.load_from_dict(stanza_dict)
    assert context.sentences[0].anchorOf == "The cat sat on the mat."
    assert context.sentences[1].anchorOf == "Felix was his name."

    word = context.sentences[1].words[0]
    assert word.uri == URIRef("https://mangosaurus.eu/rdf-data/doc_1#offset_24_29")
    assert str(word.referenceContext) == str(context)
    assert word.beginIndex == 24
    assert word.endIndex == 29
    assert word.anchorOf == "Felix"
    assert word.lemma == "Felix"
    assert word.morphofeats == [URIRef('http://purl.org/olia/olia.owl#Singular')]
    assert word.dependency[0].uri == URIRef('https://mangosaurus.eu/rdf-data/doc_1#offset_38_42')
    assert word.dependencyRelationType == 'nsubj'

def test_nif_graph():

    context = nifigator.NifContext(
        uri="https://mangosaurus.eu/rdf-data/doc_1",
        URIScheme=nifigator.OffsetBasedString,
        isString="The cat sat on the mat. Felix was his name.",
    )
    collection = nifigator.NifContextCollection(uri="https://mangosaurus.eu/rdf-data")
    collection.add_context(context)

    nlp = stanza.Pipeline("en", verbose=False)
    stanza_dict = nlp(context.isString).to_dict()
    context.load_from_dict(stanza_dict)

    g = nifigator.NifGraph(collection=collection)
    collection = g.collection

    context = collection.contexts[0]
    assert type(context) == nifigator.NifContext
    assert context.isString == "The cat sat on the mat. Felix was his name."
    assert context.uri == URIRef("https://mangosaurus.eu/rdf-data/doc_1")
    assert context.URIScheme == URIRef('http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core#OffsetBasedString')
    assert context.sentences[0].anchorOf == "The cat sat on the mat."
    assert context.sentences[1].anchorOf == "Felix was his name."
    
    word = context.sentences[1].words[0]
    assert word.uri == URIRef("https://mangosaurus.eu/rdf-data/doc_1#offset_24_29")
    assert str(word.referenceContext) == str(context)
    assert word.beginIndex == 24
    assert word.endIndex == 29
    assert word.anchorOf == "Felix"
    assert word.lemma == "Felix"
    assert word.morphofeats == [URIRef('http://purl.org/olia/olia.owl#Singular')]
    assert word.dependency[0].uri == URIRef('https://mangosaurus.eu/rdf-data/doc_1#offset_38_42')
    assert word.dependencyRelationType == 'nsubj'



"""
    Setup file for nifigator.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 4.3.1.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
from setuptools import setup

if __name__ == "__main__":
    try:
        setup(use_scm_version={"version_scheme": "no-guess-dev"})
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise
